from __future__ import annotations

from typing import TYPE_CHECKING

from hypothesis import given, settings
from hypothesis.strategies import DataObject, data
from pytest import raises

from utilities.git import (
    GetRepoRootError,
    fetch_all_tags,
    get_branch_name,
    get_ref_tags,
    get_repo_name,
    get_repo_root,
)
from utilities.hypothesis import git_repos, text_ascii

if TYPE_CHECKING:
    from pathlib import Path

    from hypothesis.strategies import DataObject


class TestFetchAllTags:
    @given(repo=git_repos())
    @settings(max_examples=1)
    def test_main(self, *, repo: Path) -> None:
        fetch_all_tags(cwd=repo)


class TestGetBranchName:
    @given(data=data(), branch=text_ascii(min_size=1))
    @settings(max_examples=1)
    def test_main(self, *, data: DataObject, branch: str) -> None:
        repo = data.draw(git_repos(branch=branch))
        result = get_branch_name(cwd=repo)
        assert result == branch


class TestGetRefTags:
    @given(repo=git_repos())
    @settings(max_examples=1)
    def test_main(self, *, repo: Path) -> None:
        tags = get_ref_tags("HEAD", cwd=repo)
        assert len(tags) == 0


class TestGetRepoName:
    @given(data=data(), name=text_ascii(min_size=1))
    @settings(max_examples=1)
    def test_main(self, *, data: DataObject, name: str) -> None:
        remote = f"https://localhost/{name}.git"
        repo = data.draw(git_repos(remote=remote))
        result = get_repo_name(cwd=repo)
        assert result == name


class TestGetRepoRoot:
    @given(repo=git_repos())
    @settings(max_examples=1)
    def test_main(self, *, repo: Path) -> None:
        root = get_repo_root(cwd=repo)
        expected = repo.resolve()
        assert root == expected

    def test_error(self, *, tmp_path: Path) -> None:
        with raises(
            GetRepoRootError, match="Path is not part of a `git` repository: .*"
        ):
            _ = get_repo_root(cwd=tmp_path)



from __future__ import annotations

from asyncio import Queue, sleep
from os import getpid
from re import search
from typing import TYPE_CHECKING, Any

from hypothesis import HealthCheck, Phase, given, settings
from hypothesis.strategies import (
    DataObject,
    DrawFn,
    binary,
    booleans,
    composite,
    data,
    dictionaries,
    lists,
    sampled_from,
    uuids,
)
from pytest import LogCaptureFixture, mark, param, raises
from redis.asyncio import Redis
from redis.asyncio.client import PubSub

from tests.conftest import SKIPIF_CI_AND_NOT_LINUX
from tests.test_operator import make_objects
from utilities.asyncio import get_items_nowait
from utilities.datetime import serialize_compact
from utilities.hypothesis import (
    int64s,
    pairs,
    settings_with_reduced_examples,
    text_ascii,
    yield_test_redis,
)
from utilities.iterables import one
from utilities.operator import is_equal
from utilities.orjson import deserialize, serialize
from utilities.redis import (
    Publisher,
    PublisherError,
    PublishError,
    PublishService,
    SubscribeService,
    _is_message,
    _RedisMessage,
    publish,
    redis_hash_map_key,
    redis_key,
    subscribe,
    yield_pubsub,
    yield_redis,
)
from utilities.sentinel import SENTINEL_REPR, Sentinel, sentinel
from utilities.tzlocal import get_now_local

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from pathlib import Path


_PUB_SUB_SLEEP = 0.1


@composite
def channels(draw: DrawFn, /) -> str:
    now = serialize_compact(get_now_local())
    key = draw(uuids())
    pid = getpid()
    return f"test_{now}_{key}_{pid}"


class TestIsMessage:
    @mark.parametrize(
        ("message", "channels", "expected"),
        [
            param(
                {
                    "type": "message",
                    "pattern": None,
                    "channel": b"channel",
                    "data": b"data",
                },
                [b"channel"],
                True,
            ),
            param(None, [], False),
            param({"type": "invalid"}, [], False),
            param({"type": "message"}, [], False),
            param({"type": "message", "pattern": False}, [], False),
            param({"type": "message", "pattern": None}, [], False),
            param(
                {"type": "message", "pattern": None, "channel": b"channel1"},
                [b"channel2"],
                False,
            ),
            param(
                {"type": "message", "pattern": None, "channel": b"channel"},
                [b"channel"],
                False,
            ),
            param(
                {
                    "type": "message",
                    "pattern": None,
                    "channel": b"channel",
                    "data": None,
                },
                [b"channel"],
                False,
            ),
        ],
    )
    def test_main(
        self, *, message: Any, channels: Sequence[bytes], expected: bool
    ) -> None:
        result = _is_message(message, channels=channels)
        assert result is expected


class TestPublish:
    @given(channel=channels(), data=lists(binary(min_size=1), min_size=1, max_size=5))
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_bytes(self, *, data: Sequence[bytes], channel: str) -> None:
        queue: Queue[bytes] = Queue()
        async with (
            yield_redis() as redis,
            subscribe(redis, channel, queue, output="bytes"),
        ):
            await sleep(_PUB_SUB_SLEEP)
            for datum in data:
                _ = await publish(redis, channel, datum)
            await sleep(_PUB_SUB_SLEEP)  # keep in context
        assert queue.qsize() == len(data)
        results = get_items_nowait(queue)
        for result, datum in zip(results, data, strict=True):
            assert isinstance(result, bytes)
            assert result == datum

    @given(channel=channels(), objects=lists(make_objects(), min_size=1, max_size=5))
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_serializer(self, *, channel: str, objects: Sequence[Any]) -> None:
        queue: Queue[Any] = Queue()
        async with (
            yield_redis() as redis,
            subscribe(redis, channel, queue, output=deserialize),
        ):
            await sleep(_PUB_SUB_SLEEP)
            for obj in objects:
                _ = await publish(redis, channel, obj, serializer=serialize)
            await sleep(_PUB_SUB_SLEEP)  # keep in context
        assert queue.qsize() == len(objects)
        results = get_items_nowait(queue)
        for result, obj in zip(results, objects, strict=True):
            assert is_equal(result, obj)

    @given(
        channel=channels(),
        messages=lists(text_ascii(min_size=1), min_size=1, max_size=5),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_text(self, *, channel: str, messages: Sequence[str]) -> None:
        queue: Queue[str] = Queue()
        async with yield_redis() as redis, subscribe(redis, channel, queue):
            await sleep(_PUB_SUB_SLEEP)
            for message in messages:
                _ = await publish(redis, channel, message)
            await sleep(_PUB_SUB_SLEEP)  # keep in context
        assert queue.qsize() == len(messages)
        results = get_items_nowait(queue)
        for result, message in zip(results, messages, strict=True):
            assert isinstance(result, str)
            assert result == message

    async def test_error(self) -> None:
        async with yield_redis() as redis:
            with raises(
                PublishError, match="Unable to publish data None with serializer None"
            ):
                _ = await publish(redis, "channel", None)


class TestPublisher:
    @given(
        channel=channels(),
        messages=lists(text_ascii(min_size=1), min_size=1, max_size=5),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_main(self, *, channel: str, messages: Sequence[str]) -> None:
        queue: Queue[str] = Queue()
        async with (
            yield_redis() as redis,
            Publisher(duration=1.0, redis=redis, sleep_core=0.1) as publisher,
            subscribe(redis, channel, queue),
        ):
            await sleep(_PUB_SUB_SLEEP)
            publisher.put_right_nowait(*((channel, m) for m in messages))
            await sleep(_PUB_SUB_SLEEP)  # keep in context
        assert queue.qsize() == len(messages)
        results = get_items_nowait(queue)
        for result, message in zip(results, messages, strict=True):
            assert isinstance(result, str)
            assert result == message

    @given(data=data())
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_error(self, *, data: DataObject) -> None:
        async with yield_test_redis(data) as test:
            publisher = Publisher(redis=test.redis)
            with raises(PublisherError, match="Error running 'Publisher'"):
                raise PublisherError(publisher=publisher)

    @given(
        channel=channels(),
        messages=lists(text_ascii(min_size=1), min_size=1, max_size=5),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_main_service(self, *, channel: str, messages: Sequence[str]) -> None:
        queue: Queue[str] = Queue()
        async with (
            yield_redis() as redis,
            PublishService(freq=0.1, timeout=1.0, redis=redis) as service,
            subscribe(redis, channel, queue),
        ):
            await sleep(_PUB_SUB_SLEEP)
            service.put_right_nowait(*((channel, m) for m in messages))
            await sleep(_PUB_SUB_SLEEP)  # keep in context
        assert queue.qsize() == len(messages)
        results = get_items_nowait(queue)
        for result, message in zip(results, messages, strict=True):
            assert isinstance(result, str)
            assert result == message


class TestRedisHashMapKey:
    @given(data=data(), key=int64s(), value=booleans())
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_get_and_set_bool(
        self, *, data: DataObject, key: int, value: bool
    ) -> None:
        async with yield_test_redis(data) as test:
            hm_key = redis_hash_map_key(test.key, int, bool)
            _ = await hm_key.set(test.redis, key, value)
            assert await hm_key.get(test.redis, key) is value

    @given(data=data(), key=booleans() | int64s(), value=booleans())
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_get_and_set_union_key(
        self, *, data: DataObject, key: bool | int, value: bool
    ) -> None:
        async with yield_test_redis(data) as test:
            hm_key = redis_hash_map_key(test.key, (bool, int), bool)
            _ = await hm_key.set(test.redis, key, value)
            assert await hm_key.get(test.redis, key) is value

    @given(data=data(), value=booleans())
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_get_and_set_sentinel_key(
        self, *, data: DataObject, value: bool
    ) -> None:
        def serializer(sentinel: Sentinel, /) -> bytes:
            return repr(sentinel).encode()

        async with yield_test_redis(data) as test:
            hm_key = redis_hash_map_key(
                test.key, Sentinel, bool, key_serializer=serializer
            )
            _ = await hm_key.set(test.redis, sentinel, value)
            assert await hm_key.get(test.redis, sentinel) is value

    @given(data=data(), key=int64s(), value=int64s() | booleans())
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_get_and_set_union_value(
        self, *, data: DataObject, key: int, value: bool | int
    ) -> None:
        async with yield_test_redis(data) as test:
            hm_key = redis_hash_map_key(test.key, int, (bool, int))
            _ = await hm_key.set(test.redis, key, value)
            assert await hm_key.get(test.redis, key) == value

    @given(data=data(), key=int64s())
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_get_and_set_sentinel_value(
        self, *, data: DataObject, key: int
    ) -> None:
        def serializer(sentinel: Sentinel, /) -> bytes:
            return repr(sentinel).encode()

        def deserializer(data: bytes, /) -> Sentinel:
            assert data == SENTINEL_REPR.encode()
            return sentinel

        async with yield_test_redis(data) as test:
            hm_key = redis_hash_map_key(
                test.key,
                int,
                Sentinel,
                value_serializer=serializer,
                value_deserializer=deserializer,
            )
            _ = await hm_key.set(test.redis, key, sentinel)
            assert await hm_key.get(test.redis, key) is sentinel

    @given(data=data(), mapping=dictionaries(int64s(), booleans()))
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_get_and_set_many(
        self, *, data: DataObject, mapping: Mapping[int, bool]
    ) -> None:
        async with yield_test_redis(data) as test:
            hm_key = redis_hash_map_key(test.key, int, bool)
            _ = await hm_key.set_many(test.redis, mapping)
            if len(mapping) == 0:
                keys = []
            else:
                keys = data.draw(lists(sampled_from(list(mapping))))
            expected = [mapping[k] for k in keys]
            assert await hm_key.get_many(test.redis, keys) == expected

    @given(data=data(), key=int64s(), value=booleans())
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_delete(self, *, data: DataObject, key: int, value: bool) -> None:
        async with yield_test_redis(data) as test:
            hm_key = redis_hash_map_key(test.key, int, bool)
            _ = await hm_key.set(test.redis, key, value)
            assert await hm_key.get(test.redis, key) is value
            _ = await hm_key.delete(test.redis, key)
            with raises(KeyError):
                _ = await hm_key.get(test.redis, key)

    @given(data=data(), key=pairs(int64s()), value=booleans())
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_delete_compound(
        self, *, data: DataObject, key: tuple[int, int], value: bool
    ) -> None:
        async with yield_test_redis(data) as test:
            hm_key = redis_hash_map_key(test.key, tuple[int, int], bool)
            _ = await hm_key.set(test.redis, key, value)
            assert await hm_key.get(test.redis, key) is value
            _ = await hm_key.delete(test.redis, key)
            with raises(KeyError):
                _ = await hm_key.get(test.redis, key)

    @given(data=data(), key=int64s(), value=booleans())
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_exists(self, *, data: DataObject, key: int, value: bool) -> None:
        async with yield_test_redis(data) as test:
            hm_key = redis_hash_map_key(test.key, int, bool)
            assert not (await hm_key.exists(test.redis, key))
            _ = await hm_key.set(test.redis, key, value)
            assert await hm_key.exists(test.redis, key)

    @given(data=data(), key=pairs(int64s()), value=booleans())
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_exists_compound(
        self, *, data: DataObject, key: tuple[int, int], value: bool
    ) -> None:
        async with yield_test_redis(data) as test:
            hm_key = redis_hash_map_key(test.key, tuple[int, int], bool)
            assert not (await hm_key.exists(test.redis, key))
            _ = await hm_key.set(test.redis, key, value)
            assert await hm_key.exists(test.redis, key)

    @given(data=data(), mapping=dictionaries(int64s(), booleans()))
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_get_all(
        self, *, data: DataObject, mapping: Mapping[int, bool]
    ) -> None:
        async with yield_test_redis(data) as test:
            hm_key = redis_hash_map_key(test.key, int, bool)
            _ = await hm_key.set_many(test.redis, mapping)
            assert await hm_key.get_all(test.redis) == mapping

    @given(data=data(), mapping=dictionaries(int64s(), booleans()))
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_keys(self, *, data: DataObject, mapping: Mapping[int, bool]) -> None:
        async with yield_test_redis(data) as test:
            hm_key = redis_hash_map_key(test.key, int, bool)
            _ = await hm_key.set_many(test.redis, mapping)
            assert await hm_key.keys(test.redis) == list(mapping)

    @given(data=data(), mapping=dictionaries(int64s(), booleans()))
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_length(
        self, *, data: DataObject, mapping: Mapping[int, bool]
    ) -> None:
        async with yield_test_redis(data) as test:
            hm_key = redis_hash_map_key(test.key, int, bool)
            _ = await hm_key.set_many(test.redis, mapping)
            assert await hm_key.length(test.redis) == len(mapping)

    @given(data=data(), key=int64s(), value=booleans())
    @settings(max_examples=1, phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_ttl(self, *, data: DataObject, key: int, value: bool) -> None:
        async with yield_test_redis(data) as test:
            hm_key = redis_hash_map_key(test.key, int, bool, ttl=0.05)
            _ = await hm_key.set(test.redis, key, value)
            await sleep(0.025)  # else next line may not work
            assert await hm_key.exists(test.redis, key)
            await sleep(0.05)
            assert not await test.redis.exists(hm_key.name)

    @given(data=data(), mapping=dictionaries(int64s(), booleans()))
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_values(
        self, *, data: DataObject, mapping: Mapping[int, bool]
    ) -> None:
        async with yield_test_redis(data) as test:
            hm_key = redis_hash_map_key(test.key, int, bool)
            _ = await hm_key.set_many(test.redis, mapping)
            assert await hm_key.values(test.redis) == list(mapping.values())


class TestRedisKey:
    @given(data=data(), value=booleans())
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_get_and_set_bool(self, *, data: DataObject, value: bool) -> None:
        async with yield_test_redis(data) as test:
            key = redis_key(test.key, bool)
            _ = await key.set(test.redis, value)
            assert await key.get(test.redis) is value

    @given(data=data(), value=booleans() | int64s())
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_get_and_set_union(
        self, *, data: DataObject, value: bool | int
    ) -> None:
        async with yield_test_redis(data) as test:
            key = redis_key(test.key, (bool, int))
            _ = await key.set(test.redis, value)
            assert await key.get(test.redis) == value

    @given(data=data())
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_get_and_set_sentinel_with_serialize(
        self, *, data: DataObject
    ) -> None:
        def serializer(sentinel: Sentinel, /) -> bytes:
            return repr(sentinel).encode()

        def deserializer(data: bytes, /) -> Sentinel:
            assert data == SENTINEL_REPR.encode()
            return sentinel

        async with yield_test_redis(data) as test:
            key = redis_key(
                test.key, Sentinel, serializer=serializer, deserializer=deserializer
            )
            _ = await key.set(test.redis, sentinel)
            assert await key.get(test.redis) is sentinel

    @given(data=data(), value=booleans())
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_delete(self, *, data: DataObject, value: bool) -> None:
        async with yield_test_redis(data) as test:
            key = redis_key(test.key, bool)
            _ = await key.set(test.redis, value)
            assert await key.get(test.redis) is value
            _ = await key.delete(test.redis)
            with raises(KeyError):
                _ = await key.get(test.redis)

    @given(data=data(), value=booleans())
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_exists(self, *, data: DataObject, value: bool) -> None:
        async with yield_test_redis(data) as test:
            key = redis_key(test.key, bool)
            assert not (await key.exists(test.redis))
            _ = await key.set(test.redis, value)
            assert await key.exists(test.redis)

    @given(data=data(), value=booleans())
    @settings(max_examples=1, phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_ttl(self, *, data: DataObject, value: bool) -> None:
        async with yield_test_redis(data) as test:
            key = redis_key(test.key, bool, ttl=0.05)
            _ = await key.set(test.redis, value)
            await sleep(0.025)  # else next line may not work
            assert await key.exists(test.redis)
            await sleep(0.05)
            assert not await key.exists(test.redis)


class TestSubscribe:
    @given(
        channel=channels(), messages=lists(binary(min_size=1), min_size=1, max_size=5)
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_bytes(self, *, channel: str, messages: Sequence[bytes]) -> None:
        queue: Queue[bytes] = Queue()
        async with (
            yield_redis() as redis,
            subscribe(redis, channel, queue, output="bytes"),
        ):
            await sleep(_PUB_SUB_SLEEP)
            for message in messages:
                await redis.publish(channel, message)
            await sleep(_PUB_SUB_SLEEP)  # keep in context
        assert queue.qsize() == len(messages)
        results = get_items_nowait(queue)
        for result, message in zip(results, messages, strict=True):
            assert isinstance(result, bytes)
            assert result == message

    @given(channel=channels(), objs=lists(make_objects(), min_size=1, max_size=5))
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_deserialize(self, *, channel: str, objs: Sequence[Any]) -> None:
        queue: Queue[Any] = Queue()
        async with (
            yield_redis() as redis,
            subscribe(redis, channel, queue, output=deserialize),
        ):
            await sleep(_PUB_SUB_SLEEP)
            for obj in objs:
                await redis.publish(channel, serialize(obj))
            await sleep(_PUB_SUB_SLEEP)  # keep in context
        assert queue.qsize() == len(objs)
        results = get_items_nowait(queue)
        for result, obj in zip(results, objs, strict=True):
            assert is_equal(result, obj)

    @given(
        channel=channels(),
        messages=lists(text_ascii(min_size=1), min_size=1, max_size=5),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_raw(self, *, channel: str, messages: Sequence[str]) -> None:
        queue: Queue[_RedisMessage] = Queue()
        async with (
            yield_redis() as redis,
            subscribe(redis, channel, queue, output="raw"),
        ):
            await sleep(_PUB_SUB_SLEEP)
            for message in messages:
                await redis.publish(channel, message)
            await sleep(_PUB_SUB_SLEEP)  # keep in context
        assert queue.qsize() == len(messages)
        results = get_items_nowait(queue)
        for result, message in zip(results, messages, strict=True):
            assert isinstance(result, dict)
            assert result["type"] == "message"
            assert result["pattern"] is None
            assert result["channel"] == channel.encode()
            assert result["data"] == message.encode()

    @given(
        channel=channels(),
        messages=lists(text_ascii(min_size=1), min_size=1, max_size=5),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_text(self, *, channel: str, messages: Sequence[str]) -> None:
        queue: Queue[_RedisMessage] = Queue()
        async with (
            yield_redis() as redis,
            subscribe(redis, channel, queue, output="raw"),
        ):
            await sleep(_PUB_SUB_SLEEP)
            for message in messages:
                await redis.publish(channel, message)
            await sleep(_PUB_SUB_SLEEP)  # keep in context
        assert queue.qsize() == len(messages)
        results = get_items_nowait(queue)
        for result, message in zip(results, messages, strict=True):
            assert isinstance(result, dict)
            assert result["type"] == "message"
            assert result["pattern"] is None
            assert result["channel"] == channel.encode()
            assert result["data"] == message.encode()


class TestSubscribeService:
    @given(
        channel=channels(),
        messages=lists(text_ascii(min_size=1), min_size=1, max_size=5),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_main(self, *, channel: str, messages: list[str]) -> None:
        async with (
            yield_redis() as redis,
            SubscribeService(timeout=1.0, redis=redis, channel=channel) as service,
        ):
            await sleep(_PUB_SUB_SLEEP)
            for message in messages:
                await redis.publish(channel, message)
            await sleep(_PUB_SUB_SLEEP)  # keep in context
        assert service.qsize() == len(messages)
        results = service.get_all_nowait()
        for result, message in zip(results, messages, strict=True):
            assert isinstance(result, str)
            assert result == message

    @given(channel=channels())
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_context_manager_already_subscribing(
        self, *, channel: str, caplog: LogCaptureFixture
    ) -> None:
        async with yield_redis() as redis:
            looper = SubscribeService(
                timeout=1.0, _debug=True, redis=redis, channel=channel
            )
            async with looper, looper:
                ...
            _ = one(m for m in caplog.messages if search(": already subscribing$", m))
            _ = one(
                m
                for m in caplog.messages
                if search(": already stopped subscription$", m)
            )


class TestYieldClient:
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_main(self) -> None:
        async with yield_redis() as client:
            assert isinstance(client, Redis)


class TestYieldPubSub:
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_main(self, *, tmp_path: Path) -> None:
        channel = str(tmp_path)
        async with yield_redis() as redis, yield_pubsub(redis, channel) as pubsub:
            assert isinstance(pubsub, PubSub)



from __future__ import annotations

import re
from collections import deque
from dataclasses import dataclass
from itertools import chain
from re import IGNORECASE, Match, escape, search
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Literal, overload, override

from utilities.iterables import CheckDuplicatesError, check_duplicates, transpose
from utilities.reprlib import get_repr

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence

    from utilities.types import StrStrMapping


DEFAULT_SEPARATOR = ","


##


def parse_bool(text: str, /) -> bool:
    """Parse text into a boolean value."""
    if search(r"^(0|False|N|No|Off)$", text, flags=IGNORECASE):
        return False
    if search(r"^(1|True|Y|Yes|On)$", text, flags=IGNORECASE):
        return True
    raise ParseBoolError(text=text)


@dataclass(kw_only=True, slots=True)
class ParseBoolError(Exception):
    text: str

    @override
    def __str__(self) -> str:
        return f"Unable to parse boolean value; got {self.text!r}"


##


def parse_none(text: str, /) -> None:
    """Parse text into the None value."""
    if search(r"^(|None)$", text, flags=IGNORECASE):
        return
    raise ParseNoneError(text=text)


@dataclass(kw_only=True, slots=True)
class ParseNoneError(Exception):
    text: str

    @override
    def __str__(self) -> str:
        return f"Unable to parse null value; got {self.text!r}"


##


def repr_encode(obj: Any, /) -> bytes:
    """Return the representation of the object encoded as bytes."""
    return repr(obj).encode()


##


_ACRONYM_PATTERN = re.compile(r"([A-Z\d]+)(?=[A-Z\d]|$)")
_SPACES_PATTERN = re.compile(r"\s+")
_SPLIT_PATTERN = re.compile(r"([\-_]*[A-Z][^A-Z]*[\-_]*)")


def snake_case(text: str, /) -> str:
    """Convert text into snake case."""
    text = _SPACES_PATTERN.sub("", text)
    if not text.isupper():
        text = _ACRONYM_PATTERN.sub(_snake_case_title, text)
        text = "_".join(s for s in _SPLIT_PATTERN.split(text) if s)
    while search("__", text):
        text = text.replace("__", "_")
    return text.lower()


def _snake_case_title(match: Match[str], /) -> str:
    return match.group(0).title()


##


LIST_SEPARATOR = DEFAULT_SEPARATOR
PAIR_SEPARATOR = "="
BRACKETS = [("(", ")"), ("[", "]"), ("{", "}")]


@overload
def split_key_value_pairs(
    text: str,
    /,
    *,
    list_separator: str = DEFAULT_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = BRACKETS,
    mapping: Literal[True],
) -> StrStrMapping: ...
@overload
def split_key_value_pairs(
    text: str,
    /,
    *,
    list_separator: str = DEFAULT_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = BRACKETS,
    mapping: Literal[False] = False,
) -> Sequence[tuple[str, str]]: ...
@overload
def split_key_value_pairs(
    text: str,
    /,
    *,
    list_separator: str = DEFAULT_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = BRACKETS,
    mapping: bool = False,
) -> Sequence[tuple[str, str]] | StrStrMapping: ...
def split_key_value_pairs(
    text: str,
    /,
    *,
    list_separator: str = DEFAULT_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = BRACKETS,
    mapping: bool = False,
) -> Sequence[tuple[str, str]] | StrStrMapping:
    """Split a string into key-value pairs."""
    try:
        texts = split_str(text, separator=list_separator, brackets=brackets)
    except SplitStrError as error:
        raise _SplitKeyValuePairsSplitError(text=text, inner=error.text) from None
    try:
        pairs = [
            split_str(text_i, separator=pair_separator, brackets=brackets, n=2)
            for text_i in texts
        ]
    except SplitStrError as error:
        raise _SplitKeyValuePairsSplitError(text=text, inner=error.text) from None
    if not mapping:
        return pairs
    try:
        check_duplicates(k for k, _ in pairs)
    except CheckDuplicatesError as error:
        raise _SplitKeyValuePairsDuplicateKeysError(
            text=text, counts=error.counts
        ) from None
    return dict(pairs)


@dataclass(kw_only=True, slots=True)
class SplitKeyValuePairsError(Exception):
    text: str


@dataclass(kw_only=True, slots=True)
class _SplitKeyValuePairsSplitError(SplitKeyValuePairsError):
    inner: str

    @override
    def __str__(self) -> str:
        return f"Unable to split {self.text!r} into key-value pairs"


@dataclass(kw_only=True, slots=True)
class _SplitKeyValuePairsDuplicateKeysError(SplitKeyValuePairsError):
    counts: Mapping[str, int]

    @override
    def __str__(self) -> str:
        return f"Unable to split {self.text!r} into a mapping since there are duplicate keys; got {get_repr(self.counts)}"


##


@overload
def split_str(
    text: str,
    /,
    *,
    separator: str = DEFAULT_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = None,
    n: Literal[1],
) -> tuple[str]: ...
@overload
def split_str(
    text: str,
    /,
    *,
    separator: str = DEFAULT_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = None,
    n: Literal[2],
) -> tuple[str, str]: ...
@overload
def split_str(
    text: str,
    /,
    *,
    separator: str = DEFAULT_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = None,
    n: Literal[3],
) -> tuple[str, str, str]: ...
@overload
def split_str(
    text: str,
    /,
    *,
    separator: str = DEFAULT_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = None,
    n: Literal[4],
) -> tuple[str, str, str, str]: ...
@overload
def split_str(
    text: str,
    /,
    *,
    separator: str = DEFAULT_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = None,
    n: Literal[5],
) -> tuple[str, str, str, str, str]: ...
@overload
def split_str(
    text: str,
    /,
    *,
    separator: str = DEFAULT_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = None,
    n: int | None = None,
) -> Sequence[str]: ...
def split_str(
    text: str,
    /,
    *,
    separator: str = DEFAULT_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = None,
    n: int | None = None,
) -> Sequence[str]:
    """Split a string, with a special provision for the empty string."""
    if text == "":
        texts = []
    elif text == _escape_separator(separator=separator):
        texts = [""]
    elif brackets is None:
        texts = text.split(separator)
    else:
        texts = _split_str_brackets(text, brackets, separator=separator)
    if n is None:
        return texts
    if len(texts) != n:
        raise _SplitStrCountError(text=text, n=n, texts=texts)
    return tuple(texts)


def _split_str_brackets(
    text: str,
    brackets: Iterable[tuple[str, str]],
    /,
    *,
    separator: str = DEFAULT_SEPARATOR,
) -> Sequence[str]:
    brackets = list(brackets)
    opens, closes = transpose(brackets)
    close_to_open = {close: open_ for open_, close in brackets}

    escapes = map(escape, chain(chain.from_iterable(brackets), [separator]))
    pattern = re.compile("|".join(escapes))

    results: Sequence[str] = []
    stack: deque[tuple[str, int]] = deque()
    last = 0

    for match in pattern.finditer(text):
        token, position = match.group(), match.start()
        if token in opens:
            stack.append((token, position))
        elif token in closes:
            if len(stack) == 0:
                raise _SplitStrClosingBracketUnmatchedError(
                    text=text, token=token, position=position
                )
            open_token, open_position = stack.pop()
            if open_token != close_to_open[token]:
                raise _SplitStrClosingBracketMismatchedError(
                    text=text,
                    opening_token=open_token,
                    opening_position=open_position,
                    closing_token=token,
                    closing_position=position,
                )
        elif (token == separator) and (len(stack) == 0):
            results.append(text[last:position].strip())
            last = position + 1
    results.append(text[last:].strip())
    if len(stack) >= 1:
        token, position = stack.pop()
        raise _SplitStrOpeningBracketUnmatchedError(
            text=text, token=token, position=position
        )
    return results


@dataclass(kw_only=True, slots=True)
class SplitStrError(Exception):
    text: str


@dataclass(kw_only=True, slots=True)
class _SplitStrCountError(SplitStrError):
    n: int
    texts: Sequence[str]

    @override
    def __str__(self) -> str:
        return f"Unable to split {self.text!r} into {self.n} part(s); got {len(self.texts)}"


@dataclass(kw_only=True, slots=True)
class _SplitStrClosingBracketMismatchedError(SplitStrError):
    opening_token: str
    opening_position: int
    closing_token: str
    closing_position: int

    @override
    def __str__(self) -> str:
        return f"Unable to split {self.text!r}; got mismatched {self.opening_token!r} at position {self.opening_position} and {self.closing_token!r} at position {self.closing_position}"


@dataclass(kw_only=True, slots=True)
class _SplitStrClosingBracketUnmatchedError(SplitStrError):
    token: str
    position: int

    @override
    def __str__(self) -> str:
        return f"Unable to split {self.text!r}; got unmatched {self.token!r} at position {self.position}"


@dataclass(kw_only=True, slots=True)
class _SplitStrOpeningBracketUnmatchedError(SplitStrError):
    token: str
    position: int

    @override
    def __str__(self) -> str:
        return f"Unable to split {self.text!r}; got unmatched {self.token!r} at position {self.position}"


def join_strs(
    texts: Iterable[str], /, *, sort: bool = False, separator: str = DEFAULT_SEPARATOR
) -> str:
    """Join a collection of strings, with a special provision for the empty list."""
    texts = list(texts)
    if sort:
        texts = sorted(texts)
    if texts == []:
        return ""
    if texts == [""]:
        return _escape_separator(separator=separator)
    return separator.join(texts)


def _escape_separator(*, separator: str = DEFAULT_SEPARATOR) -> str:
    return f"\\{separator}"


##


def str_encode(obj: Any, /) -> bytes:
    """Return the string representation of the object encoded as bytes."""
    return str(obj).encode()


##


def strip_and_dedent(text: str, /, *, trailing: bool = False) -> str:
    """Strip and dedent a string."""
    result = dedent(text.strip("\n")).strip("\n")
    return f"{result}\n" if trailing else result


__all__ = [
    "BRACKETS",
    "DEFAULT_SEPARATOR",
    "LIST_SEPARATOR",
    "PAIR_SEPARATOR",
    "ParseBoolError",
    "ParseNoneError",
    "SplitKeyValuePairsError",
    "SplitStrError",
    "join_strs",
    "parse_bool",
    "parse_none",
    "repr_encode",
    "snake_case",
    "split_key_value_pairs",
    "split_str",
    "str_encode",
    "strip_and_dedent",
]



from __future__ import annotations

from asyncio import CancelledError, Event, Queue, run, sleep, timeout
from collections import deque
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from functools import partial
from itertools import chain, count
from re import search
from typing import TYPE_CHECKING, Any, ClassVar, Self, cast, override

from hypothesis import HealthCheck, Phase, assume, given, settings
from hypothesis.strategies import (
    DataObject,
    booleans,
    data,
    integers,
    just,
    lists,
    none,
    permutations,
    sampled_from,
)
from pytest import LogCaptureFixture, approx, mark, param, raises

from utilities.asyncio import (
    EnhancedQueue,
    EnhancedTaskGroup,
    InfiniteLooper,
    InfiniteQueueLooper,
    Looper,
    LooperTimeoutError,
    UniquePriorityQueue,
    UniqueQueue,
    _InfiniteLooperDefaultEventError,
    _InfiniteLooperNoSuchEventError,
    _LooperNoTaskError,
    _LooperStats,
    get_event,
    get_items,
    get_items_nowait,
    put_items,
    put_items_nowait,
    sleep_dur,
    sleep_max_dur,
    sleep_until,
    sleep_until_rounded,
    stream_command,
    timeout_dur,
)
from utilities.dataclasses import replace_non_sentinel
from utilities.datetime import (
    MILLISECOND,
    MINUTE,
    datetime_duration_to_timedelta,
    get_now,
)
from utilities.functions import get_class_name
from utilities.hypothesis import sentinels, text_ascii
from utilities.iterables import one, unique_everseen
from utilities.pytest import skipif_windows
from utilities.sentinel import Sentinel, sentinel
from utilities.timer import Timer

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Callable, Iterator

    from utilities.types import (
        Coroutine1,
        Duration,
        DurationOrEveryDuration,
        MaybeCallableEvent,
        MaybeType,
    )


class TestEnhancedQueue:
    @given(
        xs=lists(integers()),
        wait=booleans(),
        put_all=booleans(),
        get_reverse=booleans(),
    )
    async def test_left(
        self, *, xs: list[int], wait: int, put_all: bool, get_reverse: bool
    ) -> None:
        _ = assume(not ((len(xs) == 0) and wait))
        deq: deque[int] = deque()
        for x in xs:
            deq.appendleft(x)
        queue: EnhancedQueue[int] = EnhancedQueue()
        if put_all:
            if wait:
                await queue.put_left(*xs)
            else:
                queue.put_left_nowait(*xs)
        else:
            for i, x in enumerate(xs, start=1):
                if wait:
                    await queue.put_left(x)
                else:
                    queue.put_left_nowait(x)
                assert queue.qsize() == i
        assert list(deq) == xs[::-1]
        if wait:
            res = await queue.get_all(reverse=get_reverse)
        else:
            res = queue.get_all_nowait(reverse=get_reverse)
        expected = xs if get_reverse else xs[::-1]
        assert res == expected

    @given(
        xs=lists(integers()),
        wait=booleans(),
        put_all=booleans(),
        get_reverse=booleans(),
    )
    async def test_right(
        self, *, xs: list[int], wait: int, put_all: bool, get_reverse: bool
    ) -> None:
        _ = assume(not ((len(xs) == 0) and wait))
        deq: deque[int] = deque()
        for x in xs:
            deq.append(x)
        queue: EnhancedQueue[int] = EnhancedQueue()
        if put_all:
            if wait:
                await queue.put_right(*xs)
            else:
                queue.put_right_nowait(*xs)
            assert queue.qsize() == len(xs)
        else:
            for i, x in enumerate(xs, start=1):
                if wait:
                    await queue.put_right(x)
                else:
                    queue.put_right_nowait(x)
                assert queue.qsize() == i
        assert list(deq) == xs
        if wait:
            res = await queue.get_all(reverse=get_reverse)
        else:
            res = queue.get_all_nowait(reverse=get_reverse)
        expected = xs[::-1] if get_reverse else xs
        assert res == expected


class TestEnhancedTaskGroup:
    async def test_create_task_context_coroutine(self) -> None:
        flag: bool = False

        @asynccontextmanager
        async def yield_true() -> AsyncIterator[None]:
            nonlocal flag
            try:
                flag = True
                yield
            finally:
                flag = False

        assert not flag
        async with EnhancedTaskGroup(timeout=0.1) as tg:
            _ = tg.create_task_context(yield_true())
            await sleep(0.05)
            assert flag
        assert not flag

    async def test_create_task_context_looper(self) -> None:
        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            running: bool = False

            @override
            async def _initialize(self) -> None:
                self.running = True

            @override
            async def _teardown(self) -> None:
                self.running = False

        looper = Example(duration=0.1)
        assert not looper.running
        async with EnhancedTaskGroup(timeout=0.1) as tg:
            assert not looper.running
            _ = tg.create_task_context(looper)
            await sleep(0.05)
            assert looper.running
        assert not looper.running

    async def test_max_tasks_disabled(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup() as tg:
                for _ in range(10):
                    _ = tg.create_task(sleep(0.01))
        assert timer <= 0.05

    async def test_max_tasks_enabled(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup(max_tasks=2) as tg:
                for _ in range(10):
                    _ = tg.create_task(sleep(0.01))
        assert timer >= 0.05

    async def test_timeout_pass(self) -> None:
        async with EnhancedTaskGroup(timeout=0.2) as tg:
            _ = tg.create_task(sleep(0.1))

    async def test_timeout_fail(self) -> None:
        with raises(ExceptionGroup) as exc_info:
            async with EnhancedTaskGroup(timeout=0.05) as tg:
                _ = tg.create_task(sleep(0.1))
        assert len(exc_info.value.exceptions) == 1
        error = one(exc_info.value.exceptions)
        assert isinstance(error, TimeoutError)

    async def test_custom_error(self) -> None:
        class CustomError(Exception): ...

        with raises(ExceptionGroup) as exc_info:
            async with EnhancedTaskGroup(timeout=0.05, error=CustomError) as tg:
                _ = tg.create_task(sleep(0.1))
        assert len(exc_info.value.exceptions) == 1
        error = one(exc_info.value.exceptions)
        assert isinstance(error, CustomError)


class TestGetEvent:
    def test_event(self) -> None:
        event = Event()
        assert get_event(event=event) is event

    @given(event=none() | sentinels())
    def test_none_or_sentinel(self, *, event: None | Sentinel) -> None:
        assert get_event(event=event) is event

    def test_replace_non_sentinel(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            event: Event = field(default_factory=Event)

            def replace(
                self, *, event: MaybeCallableEvent | Sentinel = sentinel
            ) -> Self:
                return replace_non_sentinel(self, event=get_event(event=event))

        event1, event2, event3 = Event(), Event(), Event()
        obj = Example(event=event1)
        assert obj.event is event1
        assert obj.replace().event is event1
        assert obj.replace(event=event2).event is event2
        assert obj.replace(event=lambda: event3).event is event3

    def test_callable(self) -> None:
        event = Event()
        assert get_event(event=lambda: event) is event


class TestGetItems:
    @given(
        xs=lists(integers(), min_size=1),
        max_size=integers(1, 10) | none(),
        wait=booleans(),
    )
    async def test_main(
        self, *, xs: list[int], max_size: int | None, wait: bool
    ) -> None:
        queue: Queue[int] = Queue()
        put_items_nowait(xs, queue)
        if wait:
            result = await get_items(queue, max_size=max_size)
        else:
            result = get_items_nowait(queue, max_size=max_size)
        assert result == xs[:max_size]


class TestInfiniteLooper:
    sleep_restart_cases: ClassVar[list[Any]] = [
        param(60.0, "for 0:01:00"),
        param(MINUTE, "for 0:01:00"),
        param(("every", 60), "until next 0:01:00"),
        param(("every", MINUTE), "until next 0:01:00"),
    ]

    async def test_main_no_errors(self) -> None:
        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            counter: int = 0

            @override
            async def _initialize(self) -> None:
                self.counter = 0

            @override
            async def _core(self) -> None:
                self.counter += 1

        async with timeout(1.0), Example(sleep_core=0.05) as looper:
            pass
        assert 15 <= looper.counter <= 25

    async def test_main_with_errors(self) -> None:
        class CustomError(Exception): ...

        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            initializations: int = 0
            counter: int = 0
            teardowns: int = 0

            @override
            async def _initialize(self) -> None:
                self.initializations += 1
                self.counter = 0

            @override
            async def _core(self) -> None:
                self.counter += 1
                if self.counter >= 5:
                    raise CustomError

            @override
            async def _teardown(self) -> None:
                self.teardowns += 1

        async with timeout(1.0), Example(sleep_core=0.05, sleep_restart=0.05) as looper:
            pass
        assert 3 <= looper.initializations <= 5
        assert 0 <= looper.counter <= 5
        assert 3 <= looper.teardowns <= 5

    async def test_blacklisted_errors(self) -> None:
        class CustomError(Exception): ...

        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            counter: int = 0

            @override
            async def _core(self) -> None:
                self.counter += 1
                if self.counter >= 5:
                    raise CustomError

            @override
            def _yield_blacklisted_errors(self) -> Iterator[type[Exception]]:
                yield CustomError

        with raises(CustomError):
            async with Example(sleep_core=0.05):
                ...

    async def test_cancelled_error(self) -> None:
        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            counter: int = 0

            @override
            async def _core(self) -> None:
                self.counter += 1
                if self.counter >= 5:
                    raise CancelledError

        async with Example(sleep_core=0.05) as service:
            pass
        assert 5 <= service.counter <= 15

    async def test_duration(self) -> None:
        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            counter: int = 0

            @override
            async def _initialize(self) -> None:
                self.counter = 0

            @override
            async def _core(self) -> None:
                self.counter += 1

        async with Example(duration=1.0, sleep_core=0.05) as looper:
            pass
        assert 15 <= looper.counter <= 25

    async def test_hashable(self) -> None:
        @dataclass(kw_only=True, unsafe_hash=True)
        class Example(InfiniteLooper[None]): ...

        looper = Example(sleep_core=0.1)
        _ = hash(looper)

    async def test_nested_context_manager(self) -> None:
        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            running: bool = False

            @override
            async def _initialize(self) -> None:
                self.running = True

            @override
            async def _teardown(self) -> None:
                self.running = False

        looper = Example()
        for _ in range(2):
            assert not looper.running
            async with timeout(0.2), looper:
                assert looper.running
                async with timeout(0.1), looper:
                    assert looper.running
                assert looper.running
            assert not looper.running

    def test_repr(self) -> None:
        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            counter: int = 0

        looper = Example()
        result = repr(looper)
        expected = "TestInfiniteLooper.test_repr.<locals>.Example(counter=0)"
        assert result == expected

    @given(n=integers(10, 11))
    async def test_setting_events(self, *, n: int) -> None:
        class TrueError(Exception): ...

        class FalseError(Exception): ...

        @dataclass(kw_only=True)
        class Example(InfiniteLooper[bool]):
            counter: int = 0
            true_counter: int = 0
            false_counter: int = 0

            @override
            async def _initialize(self) -> None:
                self.counter = 0

            @override
            async def _core(self) -> None:
                self.counter += 1
                if self.counter >= n:
                    self._set_event(event=n % 2 == 0)

            @override
            def _error_upon_core(self, error: BaseException, /) -> None:
                if isinstance(error, TrueError):
                    self.true_counter += 1
                elif isinstance(error, FalseError):
                    self.false_counter += 1

            @override
            def _yield_events_and_exceptions(
                self,
            ) -> Iterator[tuple[bool, MaybeType[Exception]]]:
                yield (True, TrueError)
                yield (False, FalseError)

        async with timeout(1.0), Example(sleep_core=0.05) as looper:
            ...
        match n % 2 == 0:
            case True:
                assert looper.true_counter >= 1, looper
                assert looper.false_counter == 0
            case False:
                assert looper.true_counter == 0
                assert looper.false_counter >= 1

    async def test_whitelisted_errors(self) -> None:
        class CustomError(BaseException): ...

        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            initializations: int = 0
            counter: int = 0

            @override
            async def _initialize(self) -> None:
                self.initializations += 1
                self.counter = 0

            @override
            async def _core(self) -> None:
                self.counter += 1
                if self.counter >= 5:
                    raise CustomError

            @override
            def _yield_whitelisted_errors(self) -> Iterator[type[BaseException]]:
                yield CustomError

        async with timeout(1.0), Example(sleep_core=0.05, sleep_restart=0.05) as looper:
            ...
        assert 3 <= looper.initializations <= 7
        assert 0 <= looper.counter <= 8

    async def test_with_coroutine_self_set_event(self) -> None:
        external: int = 0

        async def inc_external(obj: Example, /) -> None:
            nonlocal external
            for _ in range(100):  # infinite doesn't work on CI
                external += 1
                obj.counter += 1
                await sleep(0.05)

        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            initializations: int = 0
            counter: int = 0

            @override
            async def _initialize(self) -> None:
                self.initializations += 1
                self.counter = 0

            @override
            async def _core(self) -> None:
                self.counter += 1
                if self.counter >= 5:
                    self._set_event()

            @override
            def _yield_coroutines(self) -> Iterator[Callable[[], Coroutine1[None]]]:
                yield partial(inc_external, self)

        async with timeout(1.0), Example(sleep_core=0.05, sleep_restart=0.05) as looper:
            ...
        assert 3 <= looper.initializations <= 7
        assert 0 <= looper.counter <= 8
        assert 13 <= external <= 22

    async def test_with_coroutine_self_error(self) -> None:
        class CustomError(Exception): ...

        async def dummy() -> None:
            _ = await Event().wait()

        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            initializations: int = 0
            counter: int = 0

            @override
            async def _initialize(self) -> None:
                self.initializations += 1
                self.counter = 0

            @override
            async def _core(self) -> None:
                self.counter += 1
                if self.counter >= 5:
                    raise CustomError

            @override
            def _yield_coroutines(self) -> Iterator[Callable[[], Coroutine1[None]]]:
                yield dummy

        async with timeout(1.0), Example(sleep_core=0.05, sleep_restart=0.05) as looper:
            ...
        assert 3 <= looper.initializations <= 5
        assert 0 <= looper.counter <= 5

    async def test_with_looper(self) -> None:
        @dataclass(kw_only=True)
        class Child(InfiniteLooper[None]):
            counter: int = 0

            @override
            async def _initialize(self) -> None:
                self.counter = 0

            @override
            async def _core(self) -> None:
                self.counter += 1

        @dataclass(kw_only=True)
        class Parent(InfiniteLooper[None]):
            counter: int = 0
            child: Child = field(init=False, repr=False)

            @override
            def __post_init__(self) -> None:
                super().__post_init__()
                self.child = Child(sleep_core=self.sleep_core)

            @override
            async def _initialize(self) -> None:
                self.counter = 0

            @override
            async def _core(self) -> None:
                self.counter += 1

            @override
            def _yield_loopers(self) -> Iterator[InfiniteLooper]:
                yield self.child

        async with timeout(1.0), Parent(sleep_core=0.05) as parent:
            ...
        assert 15 <= parent.counter <= 25
        assert 15 <= parent.child.counter <= 25

    async def test_error_default_event(self) -> None:
        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]): ...

        looper = Example()

        with raises(
            _InfiniteLooperDefaultEventError, match=r"'Example' default event error"
        ):
            raise _InfiniteLooperDefaultEventError(looper=looper)

    @given(logger=just("logger") | none())
    @mark.parametrize(("sleep_restart", "desc"), sleep_restart_cases)
    @settings(suppress_health_check={HealthCheck.function_scoped_fixture})
    async def test_error_upon_initialize(
        self,
        *,
        sleep_restart: DurationOrEveryDuration,
        desc: str,
        logger: str | None,
        caplog: LogCaptureFixture,
    ) -> None:
        class CustomError(Exception): ...

        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            @override
            async def _initialize(self) -> None:
                raise CustomError

            @override
            async def _core(self) -> None:
                raise NotImplementedError

        async with (
            timeout(1.0),
            Example(sleep_core=0.1, sleep_restart=sleep_restart, logger=logger),
        ):
            ...
        if logger is not None:
            message = caplog.messages[0]
            expected = f"'Example' encountered 'CustomError()' whilst initializing; sleeping {desc}..."
            assert message == expected

    @given(logger=just("logger") | none())
    @mark.parametrize(("sleep_restart", "desc"), sleep_restart_cases)
    @settings(suppress_health_check={HealthCheck.function_scoped_fixture})
    async def test_error_upon_core(
        self,
        *,
        sleep_restart: DurationOrEveryDuration,
        desc: str,
        logger: str | None,
        caplog: LogCaptureFixture,
    ) -> None:
        class CustomError(Exception): ...

        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            @override
            async def _core(self) -> None:
                raise CustomError

        async with (
            timeout(1.0),
            Example(sleep_core=0.1, sleep_restart=sleep_restart, logger=logger),
        ):
            ...
        if logger is not None:
            message = caplog.messages[0]
            expected = f"'Example' encountered 'CustomError()'; sleeping {desc}..."
            assert message == expected

    @given(logger=just("logger") | none())
    @mark.parametrize(("sleep_restart", "desc"), sleep_restart_cases)
    @settings(suppress_health_check={HealthCheck.function_scoped_fixture})
    async def test_error_upon_teardown(
        self,
        *,
        sleep_restart: DurationOrEveryDuration,
        desc: str,
        logger: str | None,
        caplog: LogCaptureFixture,
    ) -> None:
        class Custom1Error(Exception): ...

        class Custom2Error(Exception): ...

        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            counter: int = 0

            @override
            async def _core(self) -> None:
                self.counter += 1
                if self.counter >= 5:
                    self._set_event()

            @override
            async def _teardown(self) -> None:
                raise Custom2Error

            @override
            def _yield_events_and_exceptions(
                self,
            ) -> Iterator[tuple[None, MaybeType[Exception]]]:
                yield (None, Custom1Error)

        async with (
            timeout(1.0),
            Example(sleep_core=0.1, sleep_restart=sleep_restart, logger=logger),
        ):
            ...
        if logger is not None:
            expected = f"'Example' encountered 'Custom2Error()' whilst tearing down; sleeping {desc}..."
            assert expected in caplog.messages

    @given(logger=just("logger") | none())
    @mark.parametrize(("sleep_restart", "desc"), sleep_restart_cases)
    @settings(suppress_health_check={HealthCheck.function_scoped_fixture})
    async def test_error_group_upon_others(
        self,
        *,
        sleep_restart: DurationOrEveryDuration,
        desc: str,
        logger: str | None,
        caplog: LogCaptureFixture,
    ) -> None:
        class CustomError(Exception): ...

        async def dummy() -> None:
            for i in count():
                if i >= 5:
                    raise CustomError
                await sleep(0.05)

        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            initializations: int = 0
            counter: int = 0

            @override
            async def _initialize(self) -> None:
                self.initializations += 1
                self.counter = 0

            @override
            async def _core(self) -> None:
                self.counter += 1

            @override
            def _yield_coroutines(self) -> Iterator[Callable[[], Coroutine1[None]]]:
                yield dummy

        async with (
            timeout(1.0),
            Example(sleep_core=0.05, sleep_restart=sleep_restart, logger=logger),
        ):
            ...
        if logger is not None:
            message = caplog.messages[0]
            expected = f"""\
'Example' encountered 1 error(s):
- Error #1/1: CustomError()
Sleeping {desc}..."""
            assert message == expected

    async def test_error_no_event_found(self) -> None:
        @dataclass(kw_only=True)
        class Example(InfiniteLooper[None]):
            counter: int = 0

            @override
            async def _initialize(self) -> None:
                self.counter = 0

            @override
            async def _core(self) -> None:
                self.counter += 1
                if self.counter >= 10:
                    self._set_event(event=cast("Any", "invalid"))

        with raises(
            _InfiniteLooperNoSuchEventError,
            match="'Example' does not have an event 'invalid'",
        ):
            async with Example():
                ...


class TestInfiniteQueueLooper:
    async def test_main(self) -> None:
        @dataclass(kw_only=True)
        class Example(InfiniteQueueLooper[None, int]):
            counter: int = 0

            @override
            async def _process_queue(self) -> None:
                self.counter += len(self._queue.get_all_nowait())

        async with timeout(1.0), Example(sleep_core=0.05) as looper:
            await sleep(0.1)
            for i in range(10):
                looper.put_right_nowait(i)
                await sleep(0.05)

            assert looper.counter == 10

    @given(n=integers(1, 10))
    def test_len_and_empty(self, *, n: int) -> None:
        class Example(InfiniteQueueLooper[None, int]):
            output: set[int] = field(default_factory=set)

            @override
            async def _process_queue(self) -> None:
                self.output.update(self._queue.get_all_nowait())

        looper = Example(sleep_core=0.05)
        assert len(looper) == 0
        assert looper.empty()
        looper.put_right_nowait(*range(n))
        assert len(looper) == looper.qsize() == n
        assert not looper.empty()

    async def test_run_until_empty_no_stop(self) -> None:
        @dataclass(kw_only=True)
        class Example(InfiniteQueueLooper[None, int]):
            output: set[int] = field(default_factory=set)

            @override
            async def _process_queue(self) -> None:
                self.output.update(self._queue.get_all_nowait())

        looper = Example(sleep_core=0.05)
        looper.put_right_nowait(*range(10))
        async with timeout(1.0), looper:
            await looper.run_until_empty()

    async def test_run_until_empty_stop(self) -> None:
        @dataclass(kw_only=True)
        class Example(InfiniteQueueLooper[None, int]):
            output: set[int] = field(default_factory=set)

            @override
            async def _process_queue(self) -> None:
                self.output.update(self._queue.get_all_nowait())

        looper = Example(sleep_core=0.05)
        looper.put_right_nowait(*range(10))
        async with looper:
            await looper.run_until_empty(stop=True)
        assert looper.empty()

    @given(logger=just("logger") | none())
    @settings(suppress_health_check={HealthCheck.function_scoped_fixture})
    async def test_error_process_items(
        self, *, logger: str | None, caplog: LogCaptureFixture
    ) -> None:
        class CustomError(Exception): ...

        @dataclass(kw_only=True)
        class Example(InfiniteQueueLooper[None, int]):
            output: set[int] = field(default_factory=set)

            @override
            async def _process_queue(self) -> None:
                raise CustomError

        async with timeout(1.0), Example(sleep_core=0.05, logger=logger) as looper:
            looper.put_left_nowait(1)
        if logger is not None:
            message = caplog.messages[0]
            expected = "'Example' encountered 'CustomError()'; sleeping for 0:01:00..."
            assert message == expected


class _ExampleLooperError(Exception): ...


@dataclass(kw_only=True)
class _ExampleCounterLooper(Looper[Any]):
    freq: Duration = field(default=10 * MILLISECOND, repr=False)
    backoff: Duration = field(default=100 * MILLISECOND, repr=False)
    _debug: bool = field(default=True, repr=False)
    count: int = 0
    max_count: int = 10

    @override
    async def _initialize_core(self) -> None:
        self.count = 0

    @override
    async def core(self) -> None:
        self.count += 1
        if self.count >= self.max_count:
            raise _ExampleLooperError


@dataclass(kw_only=True)
class _ExampleOuterLooper(_ExampleCounterLooper):
    inner: _ExampleCounterLooper = field(init=False, repr=False)

    @override
    def __post_init__(self) -> None:
        super().__post_init__()
        self.inner = _ExampleCounterLooper(
            freq=self.freq / 2,
            backoff=self.backoff / 2,
            logger=self.logger,
            timeout=self.timeout,
            timeout_error=self.timeout_error,
            max_count=round(self.max_count / 2),
        )

    @override
    def _yield_sub_loopers(self) -> Iterator[Looper]:
        yield self.inner


@dataclass(kw_only=True)
class _ExampleQueueLooper(Looper[int]):
    @override
    async def core(self) -> None:
        await super().core()
        if not self.empty():
            _ = self.get_left_nowait()


class TestLooper:
    sleep_if_failure_cases: ClassVar[list[Any]] = [
        param(True, "; sleeping for .*"),
        param(False, ""),
    ]

    async def test_main_with_nothing(self) -> None:
        looper = _ExampleCounterLooper()
        async with looper:
            ...
        assert looper.stats == _LooperStats(entries=1, stops=1)

    async def test_auto_start(self) -> None:
        looper = _ExampleCounterLooper(auto_start=True)
        with raises(TimeoutError):
            async with timeout(1.0), looper:
                ...
        self._assert_stats(looper)

    async def test_auto_start_and_timeout(self) -> None:
        looper = _ExampleCounterLooper(auto_start=True, timeout=1.0)
        async with looper:
            ...
        self._assert_stats(looper, stops=1)

    async def test_await_without_task(self) -> None:
        looper = _ExampleCounterLooper()
        with raises(_LooperNoTaskError, match=".* has no running task"):
            await looper

    async def test_context_manager_already_entered(
        self, *, caplog: LogCaptureFixture
    ) -> None:
        looper = _ExampleCounterLooper(timeout=1.0)
        async with looper, looper:
            ...
        _ = one(m for m in caplog.messages if search(": already entered$", m))

    def test_empty(self) -> None:
        looper = _ExampleCounterLooper()
        assert looper.empty()
        looper.put_left_nowait(None)
        assert not looper.empty()

    async def test_empty_upon_exit(self) -> None:
        looper = _ExampleQueueLooper(freq=0.05, empty_upon_exit=True)
        looper.put_right_nowait(0)
        assert not looper.empty()
        async with timeout(1.0), looper:
            ...
        assert looper.empty()

    async def test_explicit_start(self) -> None:
        looper = _ExampleCounterLooper()
        with raises(TimeoutError):
            async with timeout(1.0), looper:
                await looper
        self._assert_stats(looper, stops=1)

    def test_get_all_nowait(self) -> None:
        looper = _ExampleCounterLooper()
        looper.put_left_nowait(None)
        items = looper.get_all_nowait()
        assert items == [None]

    async def test_initialize_already_initializing(
        self, *, caplog: LogCaptureFixture
    ) -> None:
        class Example(_ExampleCounterLooper):
            @override
            async def _initialize_core(self) -> None:
                if self._initialization_attempts == 1:
                    _ = await super().initialize(sleep_if_failure=False)
                await super()._initialize_core()

        looper = Example()
        _ = await looper.initialize(sleep_if_failure=False)
        _ = one(m for m in caplog.messages if search(": already initializing$", m))

    @mark.parametrize(("sleep_if_failure", "extra"), sleep_if_failure_cases)
    async def test_initialize_failure(
        self, *, sleep_if_failure: bool, extra: str, caplog: LogCaptureFixture
    ) -> None:
        class Example(_ExampleCounterLooper):
            @override
            async def _initialize_core(self) -> None:
                if self._initialization_attempts == 1:
                    raise _ExampleLooperError
                await super()._initialize_core()

        looper = Example()
        _ = await looper.initialize(sleep_if_failure=sleep_if_failure)
        pattern = rf": encountered _ExampleLooperError\(\) whilst initializing{extra}$"
        _ = one(m for m in caplog.messages if search(pattern, m))

    def test_len_and_qsize(self) -> None:
        looper = _ExampleCounterLooper()
        assert len(looper) == looper.qsize() == 0
        looper.put_left_nowait(None)
        assert len(looper) == looper.qsize() == 1

    def test_replace(self) -> None:
        looper = _ExampleCounterLooper().replace(freq=10.0)
        assert looper.freq == 10.0

    async def test_request_restart(self) -> None:
        class Example(_ExampleCounterLooper):
            @override
            async def core(self) -> None:
                await super().core()
                if (self._initialization_attempts >= 2) and (
                    self.count >= (self.max_count / 2)
                ):
                    self.request_restart()

        looper = Example(auto_start=True, timeout=1.0)
        async with looper:
            ...
        self._assert_stats(
            looper,
            core_successes=79,
            core_failures=1,
            initialization_successes=16,
            tear_down_successes=15,
            restart_successes=15,
            stops=1,
        )

    def test_request_restart_already_requested(
        self, *, caplog: LogCaptureFixture
    ) -> None:
        looper = _ExampleCounterLooper()
        for _ in range(2):
            looper.request_restart()
        _ = one(
            m for m in caplog.messages if search(r": already requested restart$", m)
        )

    async def test_request_stop(self) -> None:
        class Example(_ExampleCounterLooper):
            @override
            async def core(self) -> None:
                await super().core()
                if (self._initialization_attempts >= 2) and (
                    self.count >= (self.max_count / 2)
                ):
                    self.request_stop()

        looper = Example(auto_start=True, timeout=1.0)
        async with looper:
            ...
        self._assert_stats(
            looper,
            core_successes=14,
            core_failures=1,
            initialization_successes=2,
            tear_down_successes=1,
            restart_successes=1,
            stops=1,
        )

    def test_request_stop_already_requested(self, *, caplog: LogCaptureFixture) -> None:
        looper = _ExampleCounterLooper()
        for _ in range(2):
            looper.request_stop()
        _ = one(m for m in caplog.messages if search(r": already requested stop$", m))

    async def test_request_stop_when_empty(self) -> None:
        class Example(_ExampleCounterLooper):
            @override
            async def core(self) -> None:
                await super().core()
                if self.empty():
                    await self.stop()
                match self.qsize() % 2 == 0:
                    case True:
                        _ = self.get_left_nowait()
                    case False:
                        _ = self.get_right_nowait()
                self.request_stop_when_empty()

        looper = Example(auto_start=True, timeout=1.0)
        for i in range(25):
            match i % 2 == 0:
                case True:
                    looper.put_left_nowait(i)
                case False:
                    looper.put_right_nowait(i)
        async with looper:
            ...
        self._assert_stats(
            looper,
            core_successes=25,
            core_failures=2,
            initialization_successes=3,
            tear_down_successes=2,
            restart_successes=2,
            stops=1,
        )

    def test_request_stop_when_empty_already_requested(
        self, *, caplog: LogCaptureFixture
    ) -> None:
        looper = _ExampleCounterLooper()
        for _ in range(2):
            looper.request_stop_when_empty()
        _ = one(
            m
            for m in caplog.messages
            if search(r": already requested stop when empty$", m)
        )

    @mark.parametrize(("sleep_if_failure", "extra"), sleep_if_failure_cases)
    async def test_restart_failure_during_initialization(
        self, *, sleep_if_failure: bool, extra: str, caplog: LogCaptureFixture
    ) -> None:
        class Example(_ExampleCounterLooper):
            @override
            async def _initialize_core(self) -> None:
                if self._initialization_attempts == 1:
                    raise _ExampleLooperError
                await super()._initialize_core()

        looper = Example()
        await looper.restart(sleep_if_failure=sleep_if_failure)
        pattern = rf": encountered _ExampleLooperError\(\) whilst restarting \(initialize\){extra}$"
        _ = one(m for m in caplog.messages if search(pattern, m))

    @mark.parametrize(("sleep_if_failure", "extra"), sleep_if_failure_cases)
    async def test_restart_failure_during_tear_down(
        self, *, sleep_if_failure: bool, extra: str, caplog: LogCaptureFixture
    ) -> None:
        class Example(_ExampleCounterLooper):
            @override
            async def _tear_down_core(self) -> None:
                if self._tear_down_attempts == 1:
                    raise _ExampleLooperError
                await super()._tear_down_core()

        looper = Example()
        await looper.restart(sleep_if_failure=sleep_if_failure)
        pattern = rf": encountered _ExampleLooperError\(\) whilst restarting \(tear down\){extra}$"
        _ = one(m for m in caplog.messages if search(pattern, m))

    @mark.parametrize(("sleep_if_failure", "extra"), sleep_if_failure_cases)
    async def test_restart_failure_during_tear_down_and_initialization(
        self, *, sleep_if_failure: bool, extra: str, caplog: LogCaptureFixture
    ) -> None:
        class Example(_ExampleCounterLooper):
            @override
            async def _initialize_core(self) -> None:
                if self._initialization_attempts == 1:
                    raise _ExampleLooperError
                await super()._initialize_core()

            @override
            async def _tear_down_core(self) -> None:
                if self._tear_down_attempts == 1:
                    raise _ExampleLooperError
                await super()._tear_down_core()

        looper = Example()
        await looper.restart(sleep_if_failure=sleep_if_failure)
        pattern = rf": encountered _ExampleLooperError\(\) \(tear down\) and then _ExampleLooperError\(\) \(initialization\) whilst restarting{extra}$"
        _ = one(m for m in caplog.messages if search(pattern, m))

    @mark.parametrize("n", [param(0), param(1), param(2)])
    async def test_run_until_empty(self, *, n: int) -> None:
        looper = _ExampleQueueLooper(freq=0.05)
        looper.put_right_nowait(*range(n))
        async with timeout(1.0), looper:
            await looper.run_until_empty()
        assert looper.empty()

    @mark.parametrize("auto_start", [param(True), param(False)])
    async def test_sub_looper_one(
        self, *, auto_start: bool, caplog: LogCaptureFixture
    ) -> None:
        looper = _ExampleOuterLooper(auto_start=True, timeout=1.0)
        looper.inner.auto_start = auto_start
        async with looper:
            ...
        self._assert_stats(looper, stops=1)
        self._assert_stats(
            looper.inner,
            core_successes=56,
            core_failures=13,
            initialization_successes=14,
            tear_down_successes=13,
            restart_successes=13,
            stops=1,
        )
        pattern = rf": changing sub-looper {get_class_name(_ExampleCounterLooper)}\(.*?\) to auto-start\.\.\.$"
        matches = [m for m in caplog.messages if bool(search(pattern, m))]
        if auto_start:
            assert len(matches) == 0
        else:
            _ = one(matches)

    async def test_sub_loopers_multiple(self) -> None:
        @dataclass(kw_only=True)
        class Example(_ExampleCounterLooper):
            inner1: _ExampleCounterLooper = field(init=False, repr=False)
            inner2: _ExampleCounterLooper = field(init=False, repr=False)

            @override
            def __post_init__(self) -> None:
                super().__post_init__()
                self.inner1 = _ExampleCounterLooper(
                    freq=self.freq / 2,
                    backoff=self.backoff / 2,
                    logger=self.logger,
                    timeout=self.timeout,
                    timeout_error=self.timeout_error,
                    max_count=round(self.max_count / 2),
                )
                self.inner2 = _ExampleCounterLooper(
                    freq=self.freq / 2,
                    backoff=self.backoff / 2,
                    logger=self.logger,
                    timeout=self.timeout,
                    timeout_error=self.timeout_error,
                    max_count=round(self.max_count / 2),
                )

            @override
            def _yield_sub_loopers(self) -> Iterator[Looper]:
                yield self.inner1
                yield self.inner2

        looper = Example(auto_start=True, timeout=1.0)
        async with looper:
            ...
        self._assert_stats(looper, stops=1)
        for inner in [looper.inner1, looper.inner2]:
            self._assert_stats(
                inner,
                core_successes=56,
                core_failures=13,
                initialization_successes=14,
                tear_down_successes=13,
                restart_successes=13,
                stops=1,
            )

    async def test_sub_loopers_nested(self) -> None:
        @dataclass(kw_only=True)
        class Example(_ExampleCounterLooper):
            middle: _ExampleOuterLooper = field(init=False, repr=False)

            @override
            def __post_init__(self) -> None:
                super().__post_init__()
                self.middle = _ExampleOuterLooper(
                    freq=self.freq / 2,
                    backoff=self.backoff / 2,
                    logger=self.logger,
                    timeout=self.timeout,
                    timeout_error=self.timeout_error,
                    max_count=round(self.max_count / 2),
                )

            @override
            def _yield_sub_loopers(self) -> Iterator[Looper]:
                yield self.middle

        looper = Example(auto_start=True, timeout=1.0)
        async with looper:
            ...
        self._assert_stats(looper, stops=1)
        self._assert_stats(
            looper.middle,
            core_successes=56,
            core_failures=13,
            initialization_successes=14,
            tear_down_successes=13,
            restart_successes=13,
            stops=1,
        )
        self._assert_stats(
            looper.middle.inner,
            core_successes=34,
            core_failures=34,
            initialization_successes=34,
            tear_down_successes=33,
            restart_successes=33,
            stops=1,
        )

    async def test_tear_down_already_tearing_down(
        self, *, caplog: LogCaptureFixture
    ) -> None:
        class Example(_ExampleCounterLooper):
            @override
            async def _tear_down_core(self) -> None:
                if self._tear_down_attempts == 1:
                    _ = await super().tear_down(sleep_if_failure=False)
                await super()._tear_down_core()

        looper = Example()
        _ = await looper.tear_down(sleep_if_failure=False)
        _ = one(m for m in caplog.messages if search(": already tearing down$", m))

    @mark.parametrize(("sleep_if_failure", "extra"), sleep_if_failure_cases)
    async def test_tear_down_failure(
        self, *, sleep_if_failure: bool, extra: str, caplog: LogCaptureFixture
    ) -> None:
        class Example(_ExampleCounterLooper):
            @override
            async def _tear_down_core(self) -> None:
                if self._tear_down_attempts == 1:
                    raise _ExampleLooperError
                await super()._tear_down_core()

        looper = Example()
        _ = await looper.tear_down(sleep_if_failure=sleep_if_failure)
        pattern = rf": encountered _ExampleLooperError\(\) whilst tearing down{extra}$"
        _ = one(m for m in caplog.messages if search(pattern, m))

    async def test_timeout(self) -> None:
        looper = _ExampleCounterLooper(timeout=1.0)
        async with looper:
            with raises(LooperTimeoutError, match="Timeout"):
                await looper
        self._assert_stats(looper, stops=1)

    def test_with_auto_start(self) -> None:
        looper = _ExampleCounterLooper()
        assert not looper.auto_start
        assert looper.with_auto_start.auto_start

    def _assert_stats(
        self,
        looper: _ExampleCounterLooper,
        /,
        *,
        entries: int = 1,
        core_successes: int = 45,
        core_failures: int = 5,
        initialization_successes: int = 6,
        initialization_failures: int = 0,
        tear_down_successes: int = 5,
        tear_down_failures: int = 0,
        restart_successes: int = 5,
        restart_failures: int = 0,
        stops: int = 0,
    ) -> None:
        stats = looper.stats
        assert stats.entries == entries
        assert stats.core_attempts == (stats.core_successes + stats.core_failures)
        assert stats.core_successes == approx(core_successes, rel=0.75)
        assert stats.core_failures == approx(core_failures, rel=0.75)
        assert stats.initialization_attempts == (
            stats.initialization_successes + stats.initialization_failures
        )
        assert stats.initialization_successes == approx(
            initialization_successes, rel=0.75
        )
        assert stats.initialization_failures == approx(
            initialization_failures, rel=0.75
        )
        assert stats.tear_down_attempts == (
            stats.tear_down_successes + stats.tear_down_failures
        )
        assert stats.tear_down_successes == approx(tear_down_successes, rel=0.75)
        assert stats.tear_down_failures == approx(tear_down_failures, rel=0.75)
        assert stats.restart_attempts == (
            stats.restart_successes + stats.restart_failures
        )
        assert stats.restart_successes == approx(restart_successes, rel=0.75)
        assert stats.restart_failures == approx(restart_failures, rel=0.75)
        assert stats.stops == stops


class TestPutItems:
    @given(xs=lists(integers(), min_size=1), wait=booleans())
    async def test_main(self, *, xs: list[int], wait: bool) -> None:
        queue: Queue[int] = Queue()
        if wait:
            put_items_nowait(xs, queue)
        else:
            await put_items(xs, queue)
        result: list[int] = []
        while not queue.empty():
            result.append(await queue.get())
        assert result == xs


class TestUniquePriorityQueue:
    @given(data=data(), texts=lists(text_ascii(min_size=1), min_size=1, unique=True))
    async def test_main(self, *, data: DataObject, texts: list[str]) -> None:
        items = list(enumerate(texts))
        extra = data.draw(lists(sampled_from(items)))
        items_use = data.draw(permutations(list(chain(items, extra))))
        queue: UniquePriorityQueue[int, str] = UniquePriorityQueue()
        assert queue._set == set()
        for item in items_use:
            await queue.put(item)
        assert queue._set == set(texts)
        result = await get_items(queue)
        assert result == items
        assert queue._set == set()


class TestUniqueQueue:
    @given(x=lists(integers(), min_size=1))
    async def test_main(self, *, x: list[int]) -> None:
        queue: UniqueQueue[int] = UniqueQueue()
        assert queue._set == set()
        for x_i in x:
            await queue.put(x_i)
        assert queue._set == set(x)
        result = await get_items(queue)
        expected = list(unique_everseen(x))
        assert result == expected
        assert queue._set == set()


class TestSleepDur:
    @given(duration=sampled_from([0.1, 10 * MILLISECOND]))
    @settings(phases={Phase.generate})
    async def test_main(self, *, duration: Duration) -> None:
        with Timer() as timer:
            await sleep_dur(duration=duration)
        assert timer <= datetime_duration_to_timedelta(2 * duration)

    async def test_none(self) -> None:
        with Timer() as timer:
            await sleep_dur()
        assert timer <= 0.01


class TestSleepMaxDur:
    @given(duration=sampled_from([0.1, 10 * MILLISECOND]))
    @settings(phases={Phase.generate})
    async def test_main(self, *, duration: Duration) -> None:
        with Timer() as timer:
            await sleep_max_dur(duration=duration)
        assert timer <= datetime_duration_to_timedelta(2 * duration)

    async def test_none(self) -> None:
        with Timer() as timer:
            await sleep_max_dur()
        assert timer <= 0.01


class TestSleepUntil:
    async def test_main(self) -> None:
        await sleep_until(get_now() + 10 * MILLISECOND)


class TestSleepUntilRounded:
    async def test_main(self) -> None:
        await sleep_until_rounded(10 * MILLISECOND)


class TestStreamCommand:
    @skipif_windows
    async def test_main(self) -> None:
        output = await stream_command(
            'echo "stdout message" && sleep 0.1 && echo "stderr message" >&2'
        )
        await sleep(0.01)
        assert output.return_code == 0
        assert output.stdout == "stdout message\n"
        assert output.stderr == "stderr message\n"

    @skipif_windows
    async def test_error(self) -> None:
        output = await stream_command("this-is-an-error")
        await sleep(0.01)
        assert output.return_code == 127
        assert output.stdout == ""
        assert search(
            r"^/bin/sh: (1: )?this-is-an-error: (command )?not found$", output.stderr
        )


class TestTimeoutDur:
    async def test_pass(self) -> None:
        async with timeout_dur(duration=0.2):
            await sleep(0.1)

    async def test_fail(self) -> None:
        with raises(TimeoutError):
            async with timeout_dur(duration=0.05):
                await sleep(0.1)

    async def test_custom_error(self) -> None:
        class CustomError(Exception): ...

        with raises(CustomError):
            async with timeout_dur(duration=0.05, error=CustomError):
                await sleep(0.1)


if __name__ == "__main__":
    _ = run(
        stream_command('echo "stdout message" && sleep 2 && echo "stderr message" >&2')
    )



from __future__ import annotations

from collections import defaultdict
from collections.abc import (
    AsyncIterator,
    Callable,
    Hashable,
    Iterable,
    Iterator,
    Sequence,
    Sized,
)
from collections.abc import Set as AbstractSet
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from functools import partial, reduce
from itertools import chain
from math import floor
from operator import ge, le
from re import search
from typing import Any, Literal, TypeGuard, TypeVar, assert_never, cast, override

from sqlalchemy import (
    URL,
    Column,
    Connection,
    Engine,
    Insert,
    PrimaryKeyConstraint,
    Selectable,
    Table,
    and_,
    case,
    insert,
    select,
    text,
)
from sqlalchemy.dialects.mssql import dialect as mssql_dialect
from sqlalchemy.dialects.mysql import dialect as mysql_dialect
from sqlalchemy.dialects.oracle import dialect as oracle_dialect
from sqlalchemy.dialects.postgresql import Insert as postgresql_Insert
from sqlalchemy.dialects.postgresql import dialect as postgresql_dialect
from sqlalchemy.dialects.postgresql import insert as postgresql_insert
from sqlalchemy.dialects.postgresql.asyncpg import PGDialect_asyncpg
from sqlalchemy.dialects.sqlite import Insert as sqlite_Insert
from sqlalchemy.dialects.sqlite import dialect as sqlite_dialect
from sqlalchemy.dialects.sqlite import insert as sqlite_insert
from sqlalchemy.exc import ArgumentError, DatabaseError
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine
from sqlalchemy.ext.asyncio import create_async_engine as _create_async_engine
from sqlalchemy.orm import (
    DeclarativeBase,
    InstrumentedAttribute,
    class_mapper,
    declared_attr,
)
from sqlalchemy.orm.exc import UnmappedClassError
from sqlalchemy.pool import NullPool, Pool

from utilities.asyncio import InfiniteQueueLooper, Looper, timeout_dur
from utilities.datetime import SECOND
from utilities.functions import (
    ensure_str,
    get_class_name,
    is_sequence_of_tuple_or_str_mapping,
    is_string_mapping,
    is_tuple,
    is_tuple_or_str_mapping,
)
from utilities.iterables import (
    CheckLengthError,
    CheckSubSetError,
    OneEmptyError,
    OneNonUniqueError,
    check_length,
    check_subset,
    chunked,
    merge_sets,
    merge_str_mappings,
    one,
)
from utilities.reprlib import get_repr
from utilities.text import snake_case
from utilities.types import (
    Duration,
    MaybeIterable,
    MaybeType,
    StrMapping,
    TupleOrStrMapping,
)

_T = TypeVar("_T")
type _EngineOrConnectionOrAsync = Engine | Connection | AsyncEngine | AsyncConnection
type Dialect = Literal["mssql", "mysql", "oracle", "postgresql", "sqlite"]
type ORMInstOrClass = DeclarativeBase | type[DeclarativeBase]
type TableOrORMInstOrClass = Table | ORMInstOrClass
CHUNK_SIZE_FRAC = 0.95


##


async def check_engine(
    engine: AsyncEngine,
    /,
    *,
    timeout: Duration | None = None,
    error: type[Exception] = TimeoutError,
    num_tables: int | tuple[int, float] | None = None,
) -> None:
    """Check that an engine can connect.

    Optionally query for the number of tables, or the number of columns in
    such a table.
    """
    match _get_dialect(engine):
        case "mssql" | "mysql" | "postgresql":  # skipif-ci-and-not-linux
            query = "select * from information_schema.tables"
        case "oracle":  # pragma: no cover
            query = "select * from all_objects"
        case "sqlite":
            query = "select * from sqlite_master where type='table'"
        case _ as never:
            assert_never(never)
    statement = text(query)
    async with yield_connection(engine, timeout=timeout, error=error) as conn:
        rows = (await conn.execute(statement)).all()
    if num_tables is not None:
        try:
            check_length(rows, equal_or_approx=num_tables)
        except CheckLengthError as err:
            raise CheckEngineError(
                engine=engine, rows=err.obj, expected=num_tables
            ) from None


@dataclass(kw_only=True, slots=True)
class CheckEngineError(Exception):
    engine: AsyncEngine
    rows: Sized
    expected: int | tuple[int, float]

    @override
    def __str__(self) -> str:
        return f"{get_repr(self.engine)} must have {self.expected} table(s); got {len(self.rows)}"


##


def columnwise_max(*columns: Any) -> Any:
    """Compute the columnwise max of a number of columns."""
    return _columnwise_minmax(*columns, op=ge)


def columnwise_min(*columns: Any) -> Any:
    """Compute the columnwise min of a number of columns."""
    return _columnwise_minmax(*columns, op=le)


def _columnwise_minmax(*columns: Any, op: Callable[[Any, Any], Any]) -> Any:
    """Compute the columnwise min of a number of columns."""

    def func(x: Any, y: Any, /) -> Any:
        x_none = x.is_(None)
        y_none = y.is_(None)
        col = case(
            (and_(x_none, y_none), None),
            (and_(~x_none, y_none), x),
            (and_(x_none, ~y_none), y),
            (op(x, y), x),
            else_=y,
        )
        # try auto-label
        names = {
            value for col in [x, y] if (value := getattr(col, "name", None)) is not None
        }
        try:
            (name,) = names
        except ValueError:
            return col
        else:
            return col.label(name)

    return reduce(func, columns)


##


def create_async_engine(
    drivername: str,
    /,
    *,
    username: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    database: str | None = None,
    query: StrMapping | None = None,
    poolclass: type[Pool] | None = NullPool,
) -> AsyncEngine:
    """Create a SQLAlchemy engine."""
    if query is None:
        kwargs = {}
    else:

        def func(x: MaybeIterable[str], /) -> list[str] | str:
            return x if isinstance(x, str) else list(x)

        kwargs = {"query": {k: func(v) for k, v in query.items()}}
    url = URL.create(
        drivername,
        username=username,
        password=password,
        host=host,
        port=port,
        database=database,
        **kwargs,
    )
    return _create_async_engine(url, poolclass=poolclass)


##


async def ensure_tables_created(
    engine: AsyncEngine,
    /,
    *tables_or_orms: TableOrORMInstOrClass,
    timeout: Duration | None = None,
    error: type[Exception] = TimeoutError,
) -> None:
    """Ensure a table/set of tables is/are created."""
    tables = set(map(get_table, tables_or_orms))
    match dialect := _get_dialect(engine):
        case "mysql":  # pragma: no cover
            raise NotImplementedError(dialect)
        case "postgresql":  # skipif-ci-and-not-linux
            match = "relation .* already exists"
        case "mssql":  # pragma: no cover
            match = "There is already an object named .* in the database"
        case "oracle":  # pragma: no cover
            match = "ORA-00955: name is already used by an existing object"
        case "sqlite":
            match = "table .* already exists"
        case _ as never:
            assert_never(never)
    async with yield_connection(engine, timeout=timeout, error=error) as conn:
        for table in tables:
            try:
                await conn.run_sync(table.create)
            except DatabaseError as err:
                _ensure_tables_maybe_reraise(err, match)


async def ensure_tables_dropped(
    engine: AsyncEngine,
    *tables_or_orms: TableOrORMInstOrClass,
    timeout: Duration | None = None,
    error: type[Exception] = TimeoutError,
) -> None:
    """Ensure a table/set of tables is/are dropped."""
    tables = set(map(get_table, tables_or_orms))
    match dialect := _get_dialect(engine):
        case "mysql":  # pragma: no cover
            raise NotImplementedError(dialect)
        case "postgresql":  # skipif-ci-and-not-linux
            match = "table .* does not exist"
        case "mssql":  # pragma: no cover
            match = "Cannot drop the table .*, because it does not exist or you do not have permission"
        case "oracle":  # pragma: no cover
            match = "ORA-00942: table or view does not exist"
        case "sqlite":
            match = "no such table"
        case _ as never:
            assert_never(never)
    async with yield_connection(engine, timeout=timeout, error=error) as conn:
        for table in tables:
            try:
                await conn.run_sync(table.drop)
            except DatabaseError as err:
                _ensure_tables_maybe_reraise(err, match)


##


def get_chunk_size(
    engine_or_conn: _EngineOrConnectionOrAsync,
    /,
    *,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    scaling: float = 1.0,
) -> int:
    """Get the maximum chunk size for an engine."""
    max_params = _get_dialect_max_params(engine_or_conn)
    return max(floor(chunk_size_frac * max_params / scaling), 1)


##


def get_column_names(table_or_orm: TableOrORMInstOrClass, /) -> list[str]:
    """Get the column names from a table or ORM instance/class."""
    return [col.name for col in get_columns(table_or_orm)]


##


def get_columns(table_or_orm: TableOrORMInstOrClass, /) -> list[Column[Any]]:
    """Get the columns from a table or ORM instance/class."""
    return list(get_table(table_or_orm).columns)


##


def get_primary_key_values(orm: DeclarativeBase, /) -> tuple[Any, ...]:
    """Get a tuple of the primary key values."""
    return tuple(getattr(orm, c.name) for c in yield_primary_key_columns(orm))


##


def get_table(table_or_orm: TableOrORMInstOrClass, /) -> Table:
    """Get the table from a Table or mapped class."""
    if isinstance(table_or_orm, Table):
        return table_or_orm
    if is_orm(table_or_orm):
        return cast("Table", table_or_orm.__table__)
    raise GetTableError(obj=table_or_orm)


@dataclass(kw_only=True, slots=True)
class GetTableError(Exception):
    obj: Any

    @override
    def __str__(self) -> str:
        return f"Object {self.obj} must be a Table or mapped class; got {get_class_name(self.obj)!r}"


##


def get_table_name(table_or_orm: TableOrORMInstOrClass, /) -> str:
    """Get the table name from a Table or mapped class."""
    return get_table(table_or_orm).name


##


def hash_primary_key_values(orm: DeclarativeBase, /) -> int:
    """Compute a hash of the primary key values."""
    return hash(get_primary_key_values(orm))


##


type _PairOfTupleAndTable = tuple[tuple[Any, ...], TableOrORMInstOrClass]
type _PairOfStrMappingAndTable = tuple[StrMapping, TableOrORMInstOrClass]
type _PairOfTupleOrStrMappingAndTable = tuple[TupleOrStrMapping, TableOrORMInstOrClass]
type _PairOfSequenceOfTupleOrStrMappingAndTable = tuple[
    Sequence[TupleOrStrMapping], TableOrORMInstOrClass
]
type _InsertItem = (
    _PairOfTupleOrStrMappingAndTable
    | _PairOfSequenceOfTupleOrStrMappingAndTable
    | DeclarativeBase
    | Sequence[_PairOfTupleOrStrMappingAndTable]
    | Sequence[DeclarativeBase]
)


async def insert_items(
    engine: AsyncEngine,
    *items: _InsertItem,
    snake: bool = False,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    assume_tables_exist: bool = False,
    timeout_create: Duration | None = None,
    error_create: type[Exception] = TimeoutError,
    timeout_insert: Duration | None = None,
    error_insert: type[Exception] = TimeoutError,
) -> None:
    """Insert a set of items into a database.

    These can be one of the following:
     - pair of tuple & table/class:           (x1, x2, ...), table_cls
     - pair of dict & table/class:            {k1=v1, k2=v2, ...), table_cls
     - pair of list of tuples & table/class:  [(x11, x12, ...),
                                               (x21, x22, ...),
                                               ...], table_cls
     - pair of list of dicts & table/class:   [{k1=v11, k2=v12, ...},
                                               {k1=v21, k2=v22, ...},
                                               ...], table/class
     - list of pairs of tuple & table/class:  [((x11, x12, ...), table_cls1),
                                               ((x21, x22, ...), table_cls2),
                                               ...]
     - list of pairs of dict & table/class:   [({k1=v11, k2=v12, ...}, table_cls1),
                                               ({k1=v21, k2=v22, ...}, table_cls2),
                                               ...]
     - mapped class:                          Obj(k1=v1, k2=v2, ...)
     - list of mapped classes:                [Obj(k1=v11, k2=v12, ...),
                                               Obj(k1=v21, k2=v22, ...),
                                               ...]
    """

    def build_insert(
        table: Table, values: Iterable[StrMapping], /
    ) -> tuple[Insert, Any]:
        match _get_dialect(engine):
            case "oracle":  # pragma: no cover
                return insert(table), values
            case _:
                return insert(table).values(list(values)), None

    try:
        prepared = _prepare_insert_or_upsert_items(
            partial(_normalize_insert_item, snake=snake),
            engine,
            build_insert,
            *items,
            chunk_size_frac=chunk_size_frac,
        )
    except _PrepareInsertOrUpsertItemsError as error:
        raise InsertItemsError(item=error.item) from None
    if not assume_tables_exist:
        await ensure_tables_created(
            engine, *prepared.tables, timeout=timeout_create, error=error_create
        )
    for ins, parameters in prepared.yield_pairs():
        async with yield_connection(
            engine, timeout=timeout_insert, error=error_insert
        ) as conn:
            _ = await conn.execute(ins, parameters=parameters)


@dataclass(kw_only=True, slots=True)
class InsertItemsError(Exception):
    item: _InsertItem

    @override
    def __str__(self) -> str:
        return f"Item must be valid; got {self.item}"


##


def is_orm(obj: Any, /) -> TypeGuard[ORMInstOrClass]:
    """Check if an object is an ORM instance/class."""
    if isinstance(obj, type):
        try:
            _ = class_mapper(cast("Any", obj))
        except (ArgumentError, UnmappedClassError):
            return False
        return True
    return is_orm(type(obj))


##


def is_table_or_orm(obj: Any, /) -> TypeGuard[TableOrORMInstOrClass]:
    """Check if an object is a Table or an ORM instance/class."""
    return isinstance(obj, Table) or is_orm(obj)


##


async def migrate_data(
    table_or_orm_from: TableOrORMInstOrClass,
    engine_from: AsyncEngine,
    engine_to: AsyncEngine,
    /,
    *,
    table_or_orm_to: TableOrORMInstOrClass | None = None,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    assume_tables_exist: bool = False,
    timeout_create: Duration | None = None,
    error_create: type[Exception] = TimeoutError,
    timeout_insert: Duration | None = None,
    error_insert: type[Exception] = TimeoutError,
) -> None:
    """Migrate the contents of a table from one database to another."""
    table_from = get_table(table_or_orm_from)
    async with engine_from.begin() as conn:
        rows = (await conn.execute(select(table_from))).all()
    mappings = [dict(r._mapping) for r in rows]  # noqa: SLF001
    table_to = table_from if table_or_orm_to is None else get_table(table_or_orm_to)
    items = (mappings, table_to)
    await insert_items(
        engine_to,
        items,
        chunk_size_frac=chunk_size_frac,
        assume_tables_exist=assume_tables_exist,
        timeout_create=timeout_create,
        error_create=error_create,
        timeout_insert=timeout_insert,
        error_insert=error_insert,
    )


##


def _normalize_insert_item(
    item: _InsertItem, /, *, snake: bool = False
) -> list[_NormalizedItem]:
    """Normalize an insertion item."""
    if _is_pair_of_str_mapping_and_table(item):
        mapping, table_or_orm = item
        adjusted = _map_mapping_to_table(mapping, table_or_orm, snake=snake)
        normalized = _NormalizedItem(mapping=adjusted, table=get_table(table_or_orm))
        return [normalized]
    if _is_pair_of_tuple_and_table(item):
        tuple_, table_or_orm = item
        mapping = _tuple_to_mapping(tuple_, table_or_orm)
        return _normalize_insert_item((mapping, table_or_orm), snake=snake)
    if _is_pair_of_sequence_of_tuple_or_string_mapping_and_table(item):
        items, table_or_orm = item
        pairs = [(i, table_or_orm) for i in items]
        normalized = (_normalize_insert_item(p, snake=snake) for p in pairs)
        return list(chain.from_iterable(normalized))
    if isinstance(item, DeclarativeBase):
        mapping = _orm_inst_to_dict(item)
        return _normalize_insert_item((mapping, item), snake=snake)
    try:
        _ = iter(item)
    except TypeError:
        raise _NormalizeInsertItemError(item=item) from None
    if all(map(_is_pair_of_tuple_or_str_mapping_and_table, item)):
        seq = cast("Sequence[_PairOfTupleOrStrMappingAndTable]", item)
        normalized = (_normalize_insert_item(p, snake=snake) for p in seq)
        return list(chain.from_iterable(normalized))
    if all(map(is_orm, item)):
        seq = cast("Sequence[DeclarativeBase]", item)
        normalized = (_normalize_insert_item(p, snake=snake) for p in seq)
        return list(chain.from_iterable(normalized))
    raise _NormalizeInsertItemError(item=item)


@dataclass(kw_only=True, slots=True)
class _NormalizeInsertItemError(Exception):
    item: _InsertItem

    @override
    def __str__(self) -> str:
        return f"Item must be valid; got {self.item}"


@dataclass(kw_only=True, slots=True)
class _NormalizedItem:
    mapping: StrMapping
    table: Table


def _normalize_upsert_item(
    item: _InsertItem,
    /,
    *,
    snake: bool = False,
    selected_or_all: Literal["selected", "all"] = "selected",
) -> Iterator[_NormalizedItem]:
    """Normalize an upsert item."""
    normalized = _normalize_insert_item(item, snake=snake)
    match selected_or_all:
        case "selected":
            for norm in normalized:
                values = {k: v for k, v in norm.mapping.items() if v is not None}
                yield _NormalizedItem(mapping=values, table=norm.table)
        case "all":
            yield from normalized
        case _ as never:
            assert_never(never)


##


def selectable_to_string(
    selectable: Selectable[Any], engine_or_conn: _EngineOrConnectionOrAsync, /
) -> str:
    """Convert a selectable into a string."""
    com = selectable.compile(
        dialect=engine_or_conn.dialect, compile_kwargs={"literal_binds": True}
    )
    return str(com)


##


class TablenameMixin:
    """Mix-in for an auto-generated tablename."""

    @cast("Any", declared_attr)
    def __tablename__(cls) -> str:  # noqa: N805
        return snake_case(get_class_name(cls))


##


@dataclass(kw_only=True)
class Upserter(InfiniteQueueLooper[None, _InsertItem]):
    """Upsert a set of items to a database."""

    engine: AsyncEngine
    snake: bool = False
    selected_or_all: _SelectedOrAll = "selected"
    chunk_size_frac: float = CHUNK_SIZE_FRAC
    assume_tables_exist: bool = False
    timeout_create: Duration | None = None
    error_create: type[Exception] = TimeoutError
    timeout_insert: Duration | None = None
    error_insert: type[Exception] = TimeoutError

    @override
    async def _process_queue(self) -> None:
        items = self._queue.get_all_nowait()
        await upsert_items(
            self.engine,
            *items,
            snake=self.snake,
            selected_or_all=self.selected_or_all,
            chunk_size_frac=self.chunk_size_frac,
            assume_tables_exist=self.assume_tables_exist,
            timeout_create=self.timeout_create,
            error_create=self.error_create,
            timeout_insert=self.timeout_insert,
            error_insert=self.error_insert,
        )

    @override
    def _yield_events_and_exceptions(
        self,
    ) -> Iterator[tuple[None, MaybeType[Exception]]]:
        yield (None, UpserterError)


@dataclass(kw_only=True)
class UpserterError(Exception):
    upserter: Upserter

    @override
    def __str__(self) -> str:
        return f"Error running {get_class_name(self.upserter)!r}"


@dataclass(kw_only=True)
class UpsertService(Looper[_InsertItem]):
    """Service to upsert items to a database."""

    # base
    freq: Duration = field(default=SECOND, repr=False)
    backoff: Duration = field(default=SECOND, repr=False)
    empty_upon_exit: bool = field(default=True, repr=False)
    # self
    engine: AsyncEngine
    snake: bool = False
    selected_or_all: _SelectedOrAll = "selected"
    chunk_size_frac: float = CHUNK_SIZE_FRAC
    assume_tables_exist: bool = False
    timeout_create: Duration | None = None
    error_create: type[Exception] = TimeoutError
    timeout_insert: Duration | None = None
    error_insert: type[Exception] = TimeoutError

    @override
    async def core(self) -> None:
        await super().core()
        await upsert_items(
            self.engine,
            *self.get_all_nowait(),
            snake=self.snake,
            selected_or_all=self.selected_or_all,
            chunk_size_frac=self.chunk_size_frac,
            assume_tables_exist=self.assume_tables_exist,
            timeout_create=self.timeout_create,
            error_create=self.error_create,
            timeout_insert=self.timeout_insert,
            error_insert=self.error_insert,
        )


##


type _SelectedOrAll = Literal["selected", "all"]


async def upsert_items(
    engine: AsyncEngine,
    /,
    *items: _InsertItem,
    snake: bool = False,
    selected_or_all: _SelectedOrAll = "selected",
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    assume_tables_exist: bool = False,
    timeout_create: Duration | None = None,
    error_create: type[Exception] = TimeoutError,
    timeout_insert: Duration | None = None,
    error_insert: type[Exception] = TimeoutError,
) -> None:
    """Upsert a set of items into a database.

    These can be one of the following:
     - pair of dict & table/class:            {k1=v1, k2=v2, ...), table_cls
     - pair of list of dicts & table/class:   [{k1=v11, k2=v12, ...},
                                               {k1=v21, k2=v22, ...},
                                               ...], table/class
     - list of pairs of dict & table/class:   [({k1=v11, k2=v12, ...}, table_cls1),
                                               ({k1=v21, k2=v22, ...}, table_cls2),
                                               ...]
     - mapped class:                          Obj(k1=v1, k2=v2, ...)
     - list of mapped classes:                [Obj(k1=v11, k2=v12, ...),
                                               Obj(k1=v21, k2=v22, ...),
                                               ...]
    """

    def build_insert(
        table: Table, values: Iterable[StrMapping], /
    ) -> tuple[Insert, None]:
        ups = _upsert_items_build(
            engine, table, values, selected_or_all=selected_or_all
        )
        return ups, None

    try:
        prepared = _prepare_insert_or_upsert_items(
            partial(
                _normalize_upsert_item, snake=snake, selected_or_all=selected_or_all
            ),
            engine,
            build_insert,
            *items,
            chunk_size_frac=chunk_size_frac,
        )
    except _PrepareInsertOrUpsertItemsError as error:
        raise UpsertItemsError(item=error.item) from None
    if not assume_tables_exist:
        await ensure_tables_created(
            engine, *prepared.tables, timeout=timeout_create, error=error_create
        )
    for ups, _ in prepared.yield_pairs():
        async with yield_connection(
            engine, timeout=timeout_insert, error=error_insert
        ) as conn:
            _ = await conn.execute(ups)


def _upsert_items_build(
    engine: AsyncEngine,
    table: Table,
    values: Iterable[StrMapping],
    /,
    *,
    selected_or_all: Literal["selected", "all"] = "selected",
) -> Insert:
    values = list(values)
    keys = merge_sets(*values)
    dict_nones = dict.fromkeys(keys)
    values = [{**dict_nones, **v} for v in values]
    match _get_dialect(engine):
        case "postgresql":  # skipif-ci-and-not-linux
            insert = postgresql_insert
        case "sqlite":
            insert = sqlite_insert
        case "mssql" | "mysql" | "oracle" as dialect:  # pragma: no cover
            raise NotImplementedError(dialect)
        case _ as never:
            assert_never(never)
    ins = insert(table).values(values)
    primary_key = cast("Any", table.primary_key)
    return _upsert_items_apply_on_conflict_do_update(
        values, ins, primary_key, selected_or_all=selected_or_all
    )


def _upsert_items_apply_on_conflict_do_update(
    values: Iterable[StrMapping],
    insert: postgresql_Insert | sqlite_Insert,
    primary_key: PrimaryKeyConstraint,
    /,
    *,
    selected_or_all: Literal["selected", "all"] = "selected",
) -> Insert:
    match selected_or_all:
        case "selected":
            columns = merge_sets(*values)
        case "all":
            columns = {c.name for c in insert.excluded}
        case _ as never:
            assert_never(never)
    set_ = {c: getattr(insert.excluded, c) for c in columns}
    match insert:
        case postgresql_Insert():  # skipif-ci
            return insert.on_conflict_do_update(constraint=primary_key, set_=set_)
        case sqlite_Insert():
            return insert.on_conflict_do_update(index_elements=primary_key, set_=set_)
        case _ as never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class UpsertItemsError(Exception):
    item: _InsertItem

    @override
    def __str__(self) -> str:
        return f"Item must be valid; got {self.item}"


##


@asynccontextmanager
async def yield_connection(
    engine: AsyncEngine,
    /,
    *,
    timeout: Duration | None = None,
    error: type[Exception] = TimeoutError,
) -> AsyncIterator[AsyncConnection]:
    """Yield an async connection."""
    async with timeout_dur(duration=timeout, error=error), engine.begin() as conn:
        yield conn


##


def yield_primary_key_columns(
    obj: TableOrORMInstOrClass,
    /,
    *,
    autoincrement: bool | Literal["auto", "ignore_fk"] | None = None,
) -> Iterator[Column]:
    """Yield the primary key columns of a table."""
    table = get_table(obj)
    for column in table.primary_key:
        if (autoincrement is None) or (autoincrement == column.autoincrement):
            yield column


##


def _ensure_tables_maybe_reraise(error: DatabaseError, match: str, /) -> None:
    """Re-raise the error if it does not match the required statement."""
    if not search(match, ensure_str(one(error.args))):
        raise error  # pragma: no cover


##


def _get_dialect(engine_or_conn: _EngineOrConnectionOrAsync, /) -> Dialect:
    """Get the dialect of a database."""
    dialect = engine_or_conn.dialect
    if isinstance(dialect, mssql_dialect):  # pragma: no cover
        return "mssql"
    if isinstance(dialect, mysql_dialect):  # pragma: no cover
        return "mysql"
    if isinstance(dialect, oracle_dialect):  # pragma: no cover
        return "oracle"
    if isinstance(  # skipif-ci-and-not-linux
        dialect, postgresql_dialect | PGDialect_asyncpg
    ):
        return "postgresql"
    if isinstance(dialect, sqlite_dialect):
        return "sqlite"
    msg = f"Unknown dialect: {dialect}"  # pragma: no cover
    raise NotImplementedError(msg)  # pragma: no cover


##


def _get_dialect_max_params(
    dialect_or_engine_or_conn: Dialect | _EngineOrConnectionOrAsync, /
) -> int:
    """Get the max number of parameters of a dialect."""
    match dialect_or_engine_or_conn:
        case "mssql":  # pragma: no cover
            return 2100
        case "mysql":  # pragma: no cover
            return 65535
        case "oracle":  # pragma: no cover
            return 1000
        case "postgresql":  # skipif-ci-and-not-linux
            return 32767
        case "sqlite":
            return 100
        case (
            Engine()
            | Connection()
            | AsyncEngine()
            | AsyncConnection() as engine_or_conn
        ):
            dialect = _get_dialect(engine_or_conn)
            return _get_dialect_max_params(dialect)
        case _ as never:
            assert_never(never)


##


def _is_pair_of_sequence_of_tuple_or_string_mapping_and_table(
    obj: Any, /
) -> TypeGuard[_PairOfSequenceOfTupleOrStrMappingAndTable]:
    """Check if an object is a pair of a sequence of tuples/string mappings and a table."""
    return _is_pair_with_predicate_and_table(obj, is_sequence_of_tuple_or_str_mapping)


def _is_pair_of_str_mapping_and_table(
    obj: Any, /
) -> TypeGuard[_PairOfStrMappingAndTable]:
    """Check if an object is a pair of a string mapping and a table."""
    return _is_pair_with_predicate_and_table(obj, is_string_mapping)


def _is_pair_of_tuple_and_table(obj: Any, /) -> TypeGuard[_PairOfTupleAndTable]:
    """Check if an object is a pair of a tuple and a table."""
    return _is_pair_with_predicate_and_table(obj, is_tuple)


def _is_pair_of_tuple_or_str_mapping_and_table(
    obj: Any, /
) -> TypeGuard[_PairOfTupleOrStrMappingAndTable]:
    """Check if an object is a pair of a tuple/string mapping and a table."""
    return _is_pair_with_predicate_and_table(obj, is_tuple_or_str_mapping)


def _is_pair_with_predicate_and_table(
    obj: Any, predicate: Callable[[Any], TypeGuard[_T]], /
) -> TypeGuard[tuple[_T, TableOrORMInstOrClass]]:
    """Check if an object is pair and a table."""
    return (
        isinstance(obj, tuple)
        and (len(obj) == 2)
        and predicate(obj[0])
        and is_table_or_orm(obj[1])
    )


##


def _map_mapping_to_table(
    mapping: StrMapping, table_or_orm: TableOrORMInstOrClass, /, *, snake: bool = False
) -> StrMapping:
    """Map a mapping to a table."""
    columns = get_column_names(table_or_orm)
    if not snake:
        try:
            check_subset(mapping, columns)
        except CheckSubSetError as error:
            raise _MapMappingToTableExtraColumnsError(
                mapping=mapping, columns=columns, extra=error.extra
            ) from None
        return {k: v for k, v in mapping.items() if k in columns}
    out: dict[str, Any] = {}
    for key, value in mapping.items():
        try:
            col = one(c for c in columns if snake_case(c) == snake_case(key))
        except OneEmptyError:
            raise _MapMappingToTableSnakeMapEmptyError(
                mapping=mapping, columns=columns, key=key
            ) from None
        except OneNonUniqueError as error:
            raise _MapMappingToTableSnakeMapNonUniqueError(
                mapping=mapping,
                columns=columns,
                key=key,
                first=error.first,
                second=error.second,
            ) from None
        else:
            out[col] = value
    return out


@dataclass(kw_only=True, slots=True)
class _MapMappingToTableError(Exception):
    mapping: StrMapping
    columns: Sequence[str]


@dataclass(kw_only=True, slots=True)
class _MapMappingToTableExtraColumnsError(_MapMappingToTableError):
    extra: AbstractSet[str]

    @override
    def __str__(self) -> str:
        return f"Mapping {get_repr(self.mapping)} must be a subset of table columns {get_repr(self.columns)}; got extra {self.extra}"


@dataclass(kw_only=True, slots=True)
class _MapMappingToTableSnakeMapEmptyError(_MapMappingToTableError):
    key: str

    @override
    def __str__(self) -> str:
        return f"Mapping {get_repr(self.mapping)} must be a subset of table columns {get_repr(self.columns)}; cannot find column to map to {self.key!r} modulo snake casing"


@dataclass(kw_only=True, slots=True)
class _MapMappingToTableSnakeMapNonUniqueError(_MapMappingToTableError):
    key: str
    first: str
    second: str

    @override
    def __str__(self) -> str:
        return f"Mapping {get_repr(self.mapping)} must be a subset of table columns {get_repr(self.columns)}; found columns {self.first!r}, {self.second!r} and perhaps more to map to {self.key!r} modulo snake casing"


##


def _orm_inst_to_dict(obj: DeclarativeBase, /) -> StrMapping:
    """Map an ORM instance to a dictionary."""
    cls = type(obj)

    def is_attr(attr: str, key: str, /) -> str | None:
        if isinstance(value := getattr(cls, attr), InstrumentedAttribute) and (
            value.name == key
        ):
            return attr
        return None

    def yield_items() -> Iterator[tuple[str, Any]]:
        for key in get_column_names(cls):
            attr = one(attr for attr in dir(cls) if is_attr(attr, key) is not None)
            yield key, getattr(obj, attr)

    return dict(yield_items())


##


@dataclass(kw_only=True, slots=True)
class _PrepareInsertOrUpsertItems:
    mapping: dict[Table, list[StrMapping]] = field(default_factory=dict)
    yield_pairs: Callable[[], Iterator[tuple[Insert, Any]]]

    @property
    def tables(self) -> Sequence[Table]:
        return list(self.mapping)


def _prepare_insert_or_upsert_items(
    normalize_item: Callable[[_InsertItem], Iterable[_NormalizedItem]],
    engine: AsyncEngine,
    build_insert: Callable[[Table, Iterable[StrMapping]], tuple[Insert, Any]],
    /,
    *items: Any,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
) -> _PrepareInsertOrUpsertItems:
    """Prepare a set of insert/upsert items."""
    mapping: defaultdict[Table, list[StrMapping]] = defaultdict(list)
    lengths: set[int] = set()
    try:
        for item in items:
            for normed in normalize_item(item):
                mapping[normed.table].append(normed.mapping)
                lengths.add(len(normed.mapping))
    except _NormalizeInsertItemError as error:
        raise _PrepareInsertOrUpsertItemsError(item=error.item) from None
    merged = {
        table: _prepare_insert_or_upsert_items_merge_items(table, values)
        for table, values in mapping.items()
    }
    max_length = max(lengths, default=1)
    chunk_size = get_chunk_size(
        engine, chunk_size_frac=chunk_size_frac, scaling=max_length
    )

    def yield_pairs() -> Iterator[tuple[Insert, None]]:
        for table, values in merged.items():
            for chunk in chunked(values, chunk_size):
                yield build_insert(table, chunk)

    return _PrepareInsertOrUpsertItems(mapping=mapping, yield_pairs=yield_pairs)


@dataclass(kw_only=True, slots=True)
class _PrepareInsertOrUpsertItemsError(Exception):
    item: Any

    @override
    def __str__(self) -> str:
        return f"Item must be valid; got {self.item}"


def _prepare_insert_or_upsert_items_merge_items(
    table: Table, items: Iterable[StrMapping], /
) -> list[StrMapping]:
    columns = list(yield_primary_key_columns(table))
    col_names = [c.name for c in columns]
    cols_auto = {c.name for c in columns if c.autoincrement in {True, "auto"}}
    cols_non_auto = set(col_names) - cols_auto
    mapping: defaultdict[tuple[Hashable, ...], list[StrMapping]] = defaultdict(list)
    unchanged: list[StrMapping] = []
    for item in items:
        check_subset(cols_non_auto, item)
        has_all_auto = set(cols_auto).issubset(item)
        if has_all_auto:
            pkey = tuple(item[k] for k in col_names)
            rest: StrMapping = {k: v for k, v in item.items() if k not in col_names}
            mapping[pkey].append(rest)
        else:
            unchanged.append(item)
    merged = {k: merge_str_mappings(*v) for k, v in mapping.items()}
    return [
        dict(zip(col_names, k, strict=True)) | dict(v) for k, v in merged.items()
    ] + unchanged


##


def _tuple_to_mapping(
    values: tuple[Any, ...], table_or_orm: TableOrORMInstOrClass, /
) -> dict[str, Any]:
    columns = get_column_names(table_or_orm)
    mapping = dict(zip(columns, tuple(values), strict=False))
    return {k: v for k, v in mapping.items() if v is not None}


__all__ = [
    "CHUNK_SIZE_FRAC",
    "CheckEngineError",
    "GetTableError",
    "InsertItemsError",
    "TablenameMixin",
    "UpsertItemsError",
    "UpsertService",
    "Upserter",
    "UpserterError",
    "check_engine",
    "columnwise_max",
    "columnwise_min",
    "create_async_engine",
    "ensure_tables_created",
    "ensure_tables_dropped",
    "get_chunk_size",
    "get_column_names",
    "get_columns",
    "get_primary_key_values",
    "get_table",
    "get_table_name",
    "hash_primary_key_values",
    "insert_items",
    "is_orm",
    "is_table_or_orm",
    "migrate_data",
    "selectable_to_string",
    "upsert_items",
    "yield_connection",
    "yield_primary_key_columns",
]



from __future__ import annotations

from hmac import compare_digest
from typing import TYPE_CHECKING, Literal

from streamlit import (
    button,
    empty,
    error,
    form,
    form_submit_button,
    markdown,
    secrets,
    session_state,
    stop,
    text_input,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from streamlit.elements.lib.utils import Key
    from streamlit.runtime.state import WidgetArgs, WidgetCallback, WidgetKwargs


def centered_button(
    label: str,
    /,
    *,
    key: Key | None = None,
    help: str | None = None,  # noqa: A002
    on_click: WidgetCallback | None = None,
    args: WidgetArgs | None = None,
    kwargs: WidgetKwargs | None = None,
    type: Literal["primary", "secondary"] = "secondary",  # noqa: A002
    disabled: bool = False,
    use_container_width: bool = False,
) -> bool:
    """Create a centered button."""
    style = r"<style>.row-widget.stButton {text-align: center;}</style>"
    _ = markdown(style, unsafe_allow_html=True)
    with empty():
        return button(
            label,
            key=key,
            help=help,
            on_click=on_click,
            args=args,
            kwargs=kwargs,
            type=type,
            disabled=disabled,
            use_container_width=use_container_width,
        )


_USERNAME = "username"
_PASSWORD = "password"  # noqa: S105
_PASSWORD_CORRECT = "password_correct"  # noqa: S105


def ensure_logged_in(
    *,
    skip: bool = False,
    before_form: Callable[..., None] | None = None,
    after_form: Callable[..., None] | None = None,
) -> None:
    """Ensure the user is logged in."""
    if not (skip or _check_password(before_form=before_form, after_form=after_form)):
        stop()


def _check_password(
    *,
    before_form: Callable[..., None] | None = None,
    after_form: Callable[..., None] | None = None,
) -> bool:
    """Return `True` if the user had a correct password."""
    if session_state.get("password_correct", False):
        return True
    if before_form is not None:
        before_form()
    with form("Credentials"):
        _ = text_input("Username", key=_USERNAME)
        _ = text_input("Password", type="password", key=_PASSWORD)
        _ = form_submit_button("Log in", on_click=_password_entered)
    if after_form is not None:
        after_form()
    if _PASSWORD_CORRECT in session_state:
        _ = error("Username/password combination invalid or incorrect")
    return False


def _password_entered() -> None:
    """Check whether a password entered by the user is correct."""
    if (session_state[_USERNAME] in secrets["passwords"]) and compare_digest(
        session_state[_PASSWORD], secrets.passwords[session_state[_USERNAME]]
    ):
        session_state[_PASSWORD_CORRECT] = True
        del session_state[_PASSWORD]
        del session_state[_USERNAME]
    else:
        session_state[_PASSWORD_CORRECT] = False


__all__ = ["ensure_logged_in"]



from __future__ import annotations

from utilities.contextvars import (
    _GLOBAL_BREAKPOINT,
    global_breakpoint,
    set_global_breakpoint,
)


class TestGlobalBreakpoint:
    def test_disabled(self) -> None:
        global_breakpoint()

    def test_set(self) -> None:
        try:
            set_global_breakpoint()
        finally:
            _ = _GLOBAL_BREAKPOINT.set(False)



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import override

from utilities.iterables import OneEmptyError, OneNonUniqueError, one


def extract_group(pattern: str, text: str, /, *, flags: int = 0) -> str:
    """Extract a group.

    The regex must have 1 capture group, and this must match exactly once.
    """
    compiled = re.compile(pattern, flags=flags)
    match compiled.groups:
        case 0:
            raise _ExtractGroupNoCaptureGroupsError(pattern=pattern, text=text)
        case 1:
            matches: list[str] = compiled.findall(text)
            try:
                return one(matches)
            except OneEmptyError:
                raise _ExtractGroupNoMatchesError(pattern=pattern, text=text) from None
            except OneNonUniqueError:
                raise _ExtractGroupMultipleMatchesError(
                    pattern=pattern, text=text, matches=matches
                ) from None
        case _:
            raise _ExtractGroupMultipleCaptureGroupsError(pattern=pattern, text=text)


@dataclass(kw_only=True, slots=True)
class ExtractGroupError(Exception):
    pattern: str
    text: str


@dataclass(kw_only=True, slots=True)
class _ExtractGroupMultipleCaptureGroupsError(ExtractGroupError):
    @override
    def __str__(self) -> str:
        return f"Pattern {self.pattern} must contain exactly one capture group; it had multiple"


@dataclass(kw_only=True, slots=True)
class _ExtractGroupMultipleMatchesError(ExtractGroupError):
    matches: list[str]

    @override
    def __str__(self) -> str:
        return f"Pattern {self.pattern} must match against {self.text} exactly once; matches were {self.matches}"


@dataclass(kw_only=True, slots=True)
class _ExtractGroupNoCaptureGroupsError(ExtractGroupError):
    @override
    def __str__(self) -> str:
        return f"Pattern {self.pattern} must contain exactly one capture group; it had none".format(
            self.pattern
        )


@dataclass(kw_only=True, slots=True)
class _ExtractGroupNoMatchesError(ExtractGroupError):
    @override
    def __str__(self) -> str:
        return f"Pattern {self.pattern} must match against {self.text}"


def extract_groups(pattern: str, text: str, /, *, flags: int = 0) -> list[str]:
    """Extract multiple groups.

    The regex may have any number of capture groups, and they must collectively
    match exactly once.
    """
    compiled = re.compile(pattern, flags=flags)
    if (n_groups := compiled.groups) == 0:
        raise _ExtractGroupsNoCaptureGroupsError(pattern=pattern, text=text)
    matches: list[str] = compiled.findall(text)
    match len(matches), n_groups:
        case 0, _:
            raise _ExtractGroupsNoMatchesError(pattern=pattern, text=text)
        case 1, 1:
            return matches
        case 1, _:
            return list(one(matches))
        case _:
            raise _ExtractGroupsMultipleMatchesError(
                pattern=pattern, text=text, matches=matches
            )


@dataclass(kw_only=True, slots=True)
class ExtractGroupsError(Exception):
    pattern: str
    text: str


@dataclass(kw_only=True, slots=True)
class _ExtractGroupsMultipleMatchesError(ExtractGroupsError):
    matches: list[str]

    @override
    def __str__(self) -> str:
        return f"Pattern {self.pattern} must match against {self.text} exactly once; matches were {self.matches}"


@dataclass(kw_only=True, slots=True)
class _ExtractGroupsNoCaptureGroupsError(ExtractGroupsError):
    pattern: str
    text: str

    @override
    def __str__(self) -> str:
        return f"Pattern {self.pattern} must contain at least one capture group"


@dataclass(kw_only=True, slots=True)
class _ExtractGroupsNoMatchesError(ExtractGroupsError):
    @override
    def __str__(self) -> str:
        return f"Pattern {self.pattern} must match against {self.text}"


__all__ = ["ExtractGroupError", "ExtractGroupsError", "extract_group", "extract_groups"]



from __future__ import annotations

from dataclasses import dataclass, field
from json import dumps
from logging import getLogger
from math import isclose, nan
from pathlib import Path
from typing import TYPE_CHECKING, Self, override

from psutil import swap_memory, virtual_memory

from utilities.asyncio import Looper
from utilities.contextlib import suppress_super_object_attribute_error
from utilities.datetime import SECOND, get_now

if TYPE_CHECKING:
    import datetime as dt
    from logging import Logger

    from utilities.types import Duration, PathLike


@dataclass(kw_only=True)
class MemoryMonitorService(Looper[None]):
    """Service to monitor memory usage."""

    # base
    freq: Duration = field(default=10 * SECOND, repr=False)
    backoff: Duration = field(default=10 * SECOND, repr=False)
    # self
    console: str | None = field(default=None, repr=False)
    path: PathLike = "memory.txt"
    _console: Logger | None = field(init=False, repr=False)
    _max_age: int | None = field(default=None, init=False, repr=False)
    _path: Path = field(init=False, repr=False)

    @override
    def __post_init__(self) -> None:
        super().__post_init__()
        if self.console is not None:
            self._console = getLogger(self.console)
        self._path = Path(self.path)

    @override
    async def core(self) -> None:
        await super().core()
        memory = MemoryUsage.new()
        mapping = {
            "datetime": memory.datetime.strftime("%Y-%m-%d %H:%M:%S"),
            "virtual used (mb)": memory.virtual_used_mb,
            "virtual total (mb)": memory.virtual_total_mb,
            "virtual (%)": memory.virtual_pct,
            "swap used (mb)": memory.swap_used_mb,
            "swap total (mb)": memory.swap_total_mb,
            "swap (%)": memory.swap_pct,
        }
        ser = dumps(mapping)
        with self._path.open(mode="a") as fh:
            _ = fh.write(f"{ser}\n")
        if self._console is not None:
            self._console.info("%s", mapping)


##


@dataclass(kw_only=True)
class MemoryUsage:
    """A memory usage."""

    datetime: dt.datetime = field(default_factory=get_now)
    virtual_used: int = field(repr=False)
    virtual_used_mb: int = field(init=False)
    virtual_total: int = field(repr=False)
    virtual_total_mb: int = field(init=False)
    virtual_pct: float = field(init=False)
    swap_used: int = field(repr=False)
    swap_used_mb: int = field(init=False)
    swap_total: int = field(repr=False)
    swap_total_mb: int = field(init=False)
    swap_pct: float = field(init=False)

    def __post_init__(self) -> None:
        with suppress_super_object_attribute_error():
            super().__post_init__()  # pyright: ignore[reportAttributeAccessIssue]
        self.virtual_used_mb = self._to_mb(self.virtual_used)
        self.virtual_total_mb = self._to_mb(self.virtual_total)
        self.virtual_pct = (
            nan
            if isclose(self.virtual_total, 0.0)
            else self.virtual_used / self.virtual_total
        )
        self.swap_used_mb = self._to_mb(self.swap_used)
        self.swap_total_mb = self._to_mb(self.swap_total)
        self.swap_pct = (
            nan if isclose(self.swap_total, 0.0) else self.swap_used / self.swap_total
        )

    @classmethod
    def new(cls) -> Self:
        virtual = virtual_memory()
        virtual_total = virtual.total
        swap = swap_memory()
        return cls(
            virtual_used=virtual_total - virtual.available,
            virtual_total=virtual_total,
            swap_used=swap.used,
            swap_total=swap.total,
        )

    def _to_mb(self, bytes_: int) -> int:
        return round(bytes_ / (1024**2))


__all__ = ["MemoryMonitorService", "MemoryUsage"]
